from typing import Optional, Dict, List, Callable, Set, Tuple
from collections import defaultdict

from vfmc_core import Algorithm, StepInfo, Cube
from vfmc.insertions import Insertions, Insertion, Replacement
from vfmc.orientation import Orientation

# Possible continuations for each step
NEXT_STEPS = {
    ("", ""): [("eo", "fb"), ("eo", "rl"), ("eo", "ud")],
    ("eo", "ud"): [("dr", "fb"), ("dr", "rl")],
    ("eo", "rl"): [("dr", "ud"), ("dr", "fb")],
    ("eo", "fb"): [("dr", "ud"), ("dr", "rl")],
    ("dr", "ud"): [("htr", "ud")],
    ("dr", "rl"): [("htr", "rl")],
    ("dr", "fb"): [("htr", "fb")],
    ("htr", "ud"): [("fr", "ud"), ("finish", ""), ("insertions", "")],
    ("htr", "rl"): [("fr", "rl"), ("finish", ""), ("insertions", "")],
    ("htr", "fb"): [("fr", "fb"), ("finish", ""), ("insertions", "")],
    ("fr", "ud"): [("finish", ""), ("insertions", "")],
    ("fr", "fb"): [("finish", ""), ("insertions", "")],
    ("fr", "rl"): [("finish", ""), ("insertions", "")],
    ("finish", ""): [("insertions", "")],
    ("insertions", ""): [("insertions", "")],
}

# Continuations after solving each step
DEFAULT_NEXT_STEPS = {
    ("htr", "ud"): ("fr", "ud"),
    ("htr", "rl"): ("fr", "rl"),
    ("htr", "fb"): ("fr", "fb"),
    ("fr", "ud"): ("finish", ""),
    ("fr", "fb"): ("finish", ""),
    ("fr", "rl"): ("finish", ""),
    ("finish", ""): ("insertions", ""),
}


class PartialSolution:
    """An individual solution step"""

    def __init__(
        self,
        kind: str = "",
        variant: str = "",
        alg: Algorithm = Algorithm(""),
        previous: Optional["PartialSolution"] = None,
    ):
        self.kind = kind
        self.variant = variant
        self.step_info = StepInfo(kind, variant)
        self.previous = previous
        self.alg = alg
        self.orientation = self.default_orientation()

    def default_orientation(self):
        orientation = Orientation.default_for(self.kind, self.variant)
        if self.previous is not None:
            if self.previous.kind == "eo":
                if orientation.front not in self.previous.variant:
                    orientation = orientation.y(1)
            else:
                orientation = Orientation(
                    self.previous.orientation.top, self.previous.orientation.front
                )
        return orientation

    def append(self, alg: Algorithm) -> bool:
        if not self.allows_moves(alg):
            return False
        self.alg = self.alg.merge(alg)
        return True

    def allows_moves(self, alg: Algorithm) -> bool:
        if self.previous is None:
            return True
        return self.previous.step_info.are_moves_allowed(alg)

    def full_alg(self):
        if self.previous is not None:
            return self.previous.full_alg().merge(self.alg)
        return self.alg

    def substeps(self) -> List["PartialSolution"]:
        if self.previous is None:
            return [self]
        else:
            return self.previous.substeps() + [self]

    def is_empty(self) -> bool:
        val = self.alg.is_empty()
        if self.previous:
            val = val and self.previous.is_empty()
        return val

    def x(self, ticks: int):
        self.orientation = self.orientation.x(ticks)

    def y(self, ticks: int):
        self.orientation = self.orientation.y(ticks)

    def z(self, ticks: int):
        self.orientation = self.orientation.z(ticks)

    def alg_str(self, verbose: bool = False):
        return str(self.alg)

    def default_comment(self) -> str:
        return step_name(self.kind, self.variant)

    def clone(self) -> "PartialSolution":
        return PartialSolution.create(
            kind=self.kind,
            variant=self.variant,
            alg=self.alg,
            previous=self.previous,
        )

    def __repr__(self):
        return f"{self.alg} // ({self.full_alg().len()})"

    def __eq__(self, other):
        if self.__class__ == other.__class__:
            return (
                self.kind == other.kind
                and self.variant == other.variant
                and str(self.full_alg()) == str(other.full_alg())
            )
        else:
            return False

    def __hash__(self):
        return f"{self.kind}{self.variant}:{self.full_alg()}".__hash__()

    @staticmethod
    def create(
        kind: str = "",
        variant: str = "",
        alg: Algorithm = Algorithm(""),
        previous: Optional["PartialSolution"] = None,
    ) -> "PartialSolution":
        if kind == "insertions":
            return InsertionsStep(previous=previous)
        return PartialSolution(kind, variant, alg, previous)


class Attempt:
    """Global state of the FMC attempt"""

    def __init__(self):
        self.scramble = ""
        self.cube = Cube("")
        self.inverse = False
        self.solution = PartialSolution()
        self._saved_by_kind: Dict[str, List[PartialSolution]] = defaultdict(list)
        self._done: Set[PartialSolution] = set()
        self._comments: Dict[PartialSolution, str] = {}
        self._continuations: Dict[PartialSolution, Tuple[str, str]] = {}
        self._orientations: Dict[PartialSolution, Orientation] = {}
        self._obscured: Set[PartialSolution] = set()
        self._cube_listeners = []
        self._saved_solution_listeners = []
        self._solution_attribute_listeners = []

    def set_scramble(self, s):
        self._saved_by_kind.clear()
        self._done.clear()
        self._comments.clear()
        self._continuations.clear()
        self._orientations.clear()
        self.notify_saved_solution_listeners()
        self.scramble = s
        self.inverse = False
        self.set_solution(PartialSolution("", "", previous=None))
        self.update_cube()

    def niss(self, inverse: Optional[bool] = None):
        self.set_inverse(not self.inverse)

    def set_inverse(self, b):
        if b != self.inverse:
            self.inverse = b
            self.update_cube()

    def toggle_done(self, sol: PartialSolution):
        if sol not in self._saved_by_kind[sol.kind]:
            if sol.previous:
                self.toggle_done(sol.previous)
            return
        if sol in self._done:
            self._done.remove(sol)
        else:
            self._done.add(sol)
        self.notify_solution_attribute_listeners()

    def toggle_obscured(self, sol: PartialSolution):
        if sol in self._obscured:
            self._obscured.remove(sol)
        else:
            self._obscured.add(sol)
        self.notify_solution_attribute_listeners()

    def set_comment(self, sol: PartialSolution, s: str):
        self._comments[sol] = s
        self.notify_solution_attribute_listeners()

    def get_comment(self, sol: PartialSolution) -> Optional[str]:
        return self._comments.get(sol)

    def is_done(self, sol: PartialSolution):
        return sol in self._done

    def to_str(self, sol: PartialSolution, verbose: bool = False) -> str:
        if sol in self._obscured:
            return " ".join(["?" for i in range(sol.alg.len())])
        comment = self._comments.get(sol)
        if not comment:
            comment = sol.default_comment()
        return f"{sol.alg_str(verbose)} // {comment} ({sol.full_alg().len()})"

    def append(self, alg: Algorithm) -> bool:
        if self.inverse:
            alg = alg.on_inverse()
        if not self.solution.append(alg):
            return False
        self.update_cube()
        return True

    def back(self):
        if self.solution.previous is None:
            self.set_solution(PartialSolution())
            return

        if self.solution.alg.len() > 0:
            sol = PartialSolution.create(
                self.solution.kind,
                self.solution.variant,
                previous=self.solution.previous,
                alg=Algorithm(""),
            )
            self.set_solution(sol)
        else:
            previous = self.solution.previous
            self.solution = previous.previous or PartialSolution()
            self.advance_to(previous.kind, previous.variant)

    def possible_next_steps(self, sol: PartialSolution) -> List[Tuple[str, str]]:
        return NEXT_STEPS[(sol.kind, sol.variant)]

    def possible_steps_following(self, kind, variant) -> List[Tuple[str, str]]:
        return NEXT_STEPS[(kind, variant)]

    def advance_to(self, kind: str, variant: str):
        if (kind, variant) == (self.solution.kind, self.solution.variant):
            return
        previous = (
            self.solution
            if not self.solution.alg.is_empty()
            else self.solution.previous
        )
        if previous:
            if not previous.alg.inverse_moves():
                self.inverse = False
            else:
                self.inverse = len(previous.alg.normal_moves()) == 0
        sol = PartialSolution.create(kind, variant, previous=previous)
        self.set_solution(sol)

    def advance(self):
        next = self._continuations.get(self.solution)
        if not next:
            next = DEFAULT_NEXT_STEPS.get((self.solution.kind, self.solution.variant))
        if not next:
            next = NEXT_STEPS[(self.solution.kind, self.solution.variant)][0]
        self.advance_to(*next)

    def solutions_by_kind(self) -> Dict[str, List[PartialSolution]]:
        return self._saved_by_kind

    def solutions_for_step(self, kind: str, variant: str) -> List[PartialSolution]:
        return [s for s in self._saved_by_kind.get(kind, []) if s.variant == s.variant]

    def corner_visibility(self) -> List[Tuple[int, int, int]]:
        return self.solution.step_info.corner_visibility(self.cube)

    def edge_visibility(self) -> List[Tuple[int, int]]:
        return self.solution.step_info.edge_visibility(self.cube)

    def last_solved_step(self) -> Optional[PartialSolution]:
        for step in reversed(self.solution.substeps()):
            if step.step_info.is_solved(self.cube):
                return step
        return None

    def reset(self):
        alg = self.solution.alg
        # Clear only the moves for the current side
        if self.inverse:
            alg = Algorithm(" ".join(alg.normal_moves()))
        else:
            alg = Algorithm(f"({' '.join(alg.inverse_moves())})")
        new_solution = PartialSolution.create(
            self.solution.kind,
            self.solution.variant,
            previous=self.solution.previous,
            alg=alg,
        )
        self.set_solution(new_solution)

    def set_solution(self, sol: PartialSolution):
        self.solution = sol
        o = self._orientations.get(sol)
        if o:
            self.solution.orientation = o
        self.update_cube()
        self.notify_solution_attribute_listeners()

    def solve(self, num_solutions: int) -> List[PartialSolution]:
        """Find solutions for the current step"""
        sol = self.solution
        on_inverse = self.inverse
        self.set_inverse(False)
        existing = set(str(s) for s in self.solutions_for_step(sol.kind, sol.variant))
        algs = sol.step_info.solve(self.cube, len(existing) + num_solutions)
        solutions = []
        for alg in algs:
            base_alg = Algorithm(str(sol.alg))
            base_alg.merge(alg)
            s = PartialSolution.create(
                kind=sol.kind,
                variant=sol.variant,
                previous=sol.previous,
                alg=sol.alg.merge(alg),
            )
            if str(s) not in existing:
                solutions.append(s)
            if len(solutions) >= num_solutions:
                break
        self.set_inverse(on_inverse)
        return solutions

    def save(self) -> Optional[PartialSolution]:
        sol = self.solution.clone()
        is_solved = sol.step_info.is_solved(self.cube)
        if not is_solved and not self.get_comment(sol):
            case = self.solution.step_info.case_name(self.cube)
            comment = f"{step_name(self.solution.kind, self.solution.variant)}-{case}"
            self.set_comment(sol, comment)
        if is_solved:
            if sol.kind in {"eo", "dr"}:
                self.reset()
            else:
                self.advance()
        self.save_solutions([sol])
        self.solution
        return sol

    def save_solutions(self, sols: List[PartialSolution]):
        new_sols_by_key = defaultdict(list)

        def sort_key(sol):
            return (
                sol.full_alg().len(),
                len(sol.alg.inverse_moves()) > 0,
                len(sol.alg.normal_moves()) > 0,
                str(sol.alg),
            )

        for s in sols:
            new_sols_by_key[s.kind].append(s)
        for kind, sols_for_kind in new_sols_by_key.items():
            existing = self._saved_by_kind[kind]
            existing_algs = set(str(s.full_alg()) for s in existing)
            existing += [
                s for s in sols_for_kind if not str(s.full_alg()) in existing_algs
            ]
            existing.sort(key=sort_key)
        self.notify_saved_solution_listeners()

    def add_saved_solution_listener(self, callback: Callable):
        self._saved_solution_listeners.append(callback)

    def add_solution_attribute_listener(self, callback: Callable):
        self._solution_attribute_listeners.append(callback)

    def add_cube_listener(self, callback: Callable):
        self._cube_listeners.append(callback)

    def update_cube(self):
        self.cube = Cube(self.scramble)
        self.cube.apply(self.solution.full_alg())
        if self.inverse:
            self.cube.invert()
        self.notify_cube_listeners()

    def notify_saved_solution_listeners(self):
        for listener in self._saved_solution_listeners:
            listener()

    def notify_solution_attribute_listeners(self):
        for listener in self._solution_attribute_listeners:
            listener()

    def notify_cube_listeners(self):
        for listener in self._cube_listeners:
            listener()


class InsertionsStep(PartialSolution):
    def __init__(self, previous):
        super().__init__(kind="insertions", previous=previous)
        self.insertions = Insertions(previous.full_alg())
        self.inserted_algs: Dict[str, str] = {}
        self.alg_with_markers = str(self.insertions.original.all_on_normal())

    def insertion_symbol(self, index):
        choices = "^#*@"
        n = int(index / len(choices)) + 1
        return choices[index % len(choices)] * n

    def replacement_symbol(self, index):
        open = "[{"
        close = "]}"
        n = int(index / len(open)) + 1
        open = open[index % len(open)] * n
        close = close[index % len(close)] * n
        return f"{open}{close}"

    def clone(self) -> "PartialSolution":
        step = InsertionsStep(previous=self.previous)
        step.alg = self.alg
        step.alg_with_markers = self.alg_with_markers
        step.insertions = self.insertions
        step.inserted_algs = self.inserted_algs
        return step

    def alg_str(self, verbose: bool = False):
        lines = []
        if not verbose:
            return " ".join(self.inserted_algs.keys())

        lines += [""] + [self.alg_with_markers]
        if self.inserted_algs:
            lines += [f"{symbol} = {alg}" for symbol, alg in self.inserted_algs.items()]
            lines += [""] + [str(Algorithm("").merge(self.insertions.replacement))]
        return "\n".join(lines + [""])

    def append(self, alg: Algorithm) -> bool:
        raise ValueError("Double-click on solution to find insertions")

    def set_replacement(self, text, pos):
        self.insertions.set_replacement(text, pos)
        self.alg = self.insertions.net_alg()

    def add_markers(self):
        # Add placeholders for insertions in preceding steps
        self.alg_with_markers = ""
        self.inserted_algs.clear()
        orig = self.insertions.original.all_on_normal().normal_moves()
        final = []
        pos = 0
        n_insert = -1
        n_replace = -1
        edits = self.insertions.get_edits()
        for edit in edits:
            if isinstance(edit, Insertion):
                n_insert += 1
                symbol = self.insertion_symbol(n_insert)
                final.extend(orig[pos : edit.pos])
                pos = edit.pos
                final.append(symbol)
                self.inserted_algs[symbol] = " ".join(edit.moves)
            elif isinstance(edit, Replacement):
                n_replace += 1
                symbol = self.replacement_symbol(n_replace)
                final.extend(orig[pos : edit.start])
                pos = edit.start
                final.append(symbol[: int(len(symbol) / 2)])
                final.extend(orig[pos : edit.end])
                pos = edit.end
                final.append(symbol[int(len(symbol) / 2) :])
                self.inserted_algs[symbol] = " ".join(edit.moves)
        final.extend(orig[pos:])
        self.alg_with_markers = " ".join(final)


def step_name(kind, variant):
    s = kind
    if s not in {"htr", "fr"}:
        s = f"{s}{variant}"
    return s



import functools
import logging


def catch_errors(func):
    """
    Untrapped errors in PyQt event handlers can cause crashes
    They should use this decorator to catch and log errors
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            # Log the actual exception that was raised
            logging.exception(f"Error in event handler {func.__name__}: {str(e)}")
            return None

    return wrapper


def catch_and_return(return_value=None):
    """
    Similar to above, but specify a valid return value when an exception is caught
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                # Log the actual exception that was raised
                logging.exception(f"Error in event handler {func.__name__}: {str(e)}")
                return return_value

        return wrapper

    return decorator



import dataclasses
import os
import sys
import math
import logging
import traceback
import re
from functools import cached_property
from typing import List, Set, Optional
from importlib.metadata import version

from PyQt5.QtWidgets import (
    QApplication,
    QMainWindow,
    QVBoxLayout,
    QHBoxLayout,
    QWidget,
    QLabel,
    QLineEdit,
    QPushButton,
    QComboBox,
    QListWidget,
    QMessageBox,
    QSizePolicy,
    QStyledItemDelegate,
    QListWidgetItem,
    QAction,
    QFileDialog,
    QInputDialog,
    QDialog,
    QScrollArea,
)
from PyQt5.QtCore import Qt, QEvent
from PyQt5.QtGui import QKeySequence

from vfmc import catch_errors, catch_and_return
from vfmc.attempt import PartialSolution, Attempt, step_name
from vfmc import prefs
from vfmc.palette import Palette
from vfmc.prefs import preferences
from vfmc.viz import CubeViz, CubeWidget
from vfmc_core import Cube, Algorithm, StepInfo, scramble as gen_scramble

# Basic set of cube moves
MOVE_REGEX = r"[rRuUfFlLdDbB '2]*"

PREFERRED_AXIS = {
    ("eo", "ud"): (["fb", "rl"], ["ud"]),
    ("eo", "rl"): (["ud", "fb"], ["rl"]),
    ("eo", "fb"): (["ud", "rl"], ["fb"]),
    ("*", "ud"): (["ud"], ["fb", "rl"]),
    ("*", "fb"): (["fb"], ["ud", "rl"]),
    ("*", "rl"): (["rl"], ["fb", "ud"]),
    ("*", "*"): (["ud"], ["fb", "rl"]),
}

AXIS_ORIENTATIONS = {
    ("ud", "fb"): (0, 0, 0),
    ("ud", "rl"): (0, 0, -math.pi / 2),
    ("fb", "ud"): (math.pi / 2, 0, 0),
    ("fb", "rl"): (math.pi / 2, 0, -math.pi / 2),
    ("rl", "fb"): (0, -math.pi / 2, 0),
    ("rl", "ud"): (0, -math.pi / 2, math.pi / 2),
}

SOLUTION = Qt.UserRole
BOLD = Qt.UserRole + 1
STRIKETHROUGH = Qt.UserRole + 2


class AppWindow(QMainWindow):
    """Main window for cube exploration with PyQt"""

    def __init__(self):
        super(AppWindow, self).__init__()

        self.setWindowTitle("VFMC")
        self.resize(1200, 800)

        self.attempt = Attempt()
        self.attempt.add_saved_solution_listener(self.populate_saved_solutions)
        self.attempt.add_solution_attribute_listener(self.format_saved_solutions)
        self.previous_solution = self.attempt.solution

        self.commands = Commands(self)
        self.command_history = []  # As entered by the user
        self.history_pointer = -1  # For traversing command history via up/down keys

        self.step_order = ["eo", "dr", "htr", "fr", "finish", "insertions"]

        self._create_menus()

        # Create GUI controls
        central_widget = self._empty_container(QVBoxLayout())
        central_widget.layout().setSpacing(4)
        central_widget.layout().setContentsMargins(10, 10, 10, 10)
        self.setCentralWidget(central_widget)
        w = self._empty_container(QHBoxLayout())
        w.layout().setContentsMargins(10, 0, 10, 0)
        w.layout().setSpacing(10)
        w.layout().addWidget(self.cube_widget)
        w.layout().addWidget(self.current_solution_scroll_widget)
        r2 = self._empty_container(QHBoxLayout())
        v = self._empty_container(QVBoxLayout())
        v.layout().addWidget(self.text_input_widget)
        v.layout().addWidget(self.status_widget)
        r2.layout().addWidget(v)
        r2.layout().addWidget(self.gui_commands_widget)
        central_widget.layout().addWidget(w)
        central_widget.layout().addWidget(r2)
        central_widget.layout().addWidget(self.solutions_widget, 1)

        # Set initial scramble
        self.commands.execute("scramble")

        # Set focus to command input
        self.command_input.setFocus()

    @cached_property
    def cube_widget(self) -> QWidget:
        # Cube visualization
        cube_widget = QWidget()
        cube_widget.setLayout(QVBoxLayout())
        cube_widget.layout().setContentsMargins(0, 0, 0, 0)
        cube_widget.layout().setSpacing(0)

        # Graphical Cube
        self.viz = CubeViz(self.attempt)
        self.cube_viz_widget = CubeWidget(self.viz)
        cube_widget.layout().addWidget(self.cube_viz_widget)

        # Status labels below cube graphics
        status_container = QWidget()
        status_layout = QHBoxLayout(status_container)
        status_layout.setContentsMargins(0, 0, 0, 0)
        status_layout.setSpacing(0)

        # Left label - Step kind and variant
        step_label = QLabel("Step")
        step_label.setMinimumHeight(40)
        status_layout.addWidget(step_label, 1)  # Give it a stretch factor of 1

        # Inverse marker
        niss_label = QLabel("NISS")
        niss_label.setAlignment(Qt.AlignCenter)
        niss_label.setMinimumHeight(40)
        status_layout.addWidget(niss_label, 1)  # Give it a stretch factor of 1

        # Right label - Case name
        case_label = QLabel("Case")
        case_label.setAlignment(Qt.AlignRight)
        case_label.setMinimumHeight(40)
        status_layout.addWidget(case_label, 1)  # Give it a stretch factor of 1

        def update_bg_color():
            bg = str(hex(preferences.background_color))[2:]
            text_color = 255 if preferences.background_color < 128 else 0
            tc = str(hex(text_color))[2:]
            label_style = f"background-color: #{bg}{bg}{bg}; color: #{tc}{tc}{tc}; font-weight: bold; font-size: 18px; padding: 5px;"
            step_label.setStyleSheet(label_style)
            niss_label.setStyleSheet(label_style)
            case_label.setStyleSheet(label_style)

        update_bg_color()
        preferences.add_listener(update_bg_color)

        def refresh():
            sol = self.attempt.solution
            step_text = f"{sol.kind}{sol.variant}"
            step_label.setText(step_text)

            # Update NISS label
            niss_label.setText("(inverse)" if self.attempt.inverse else "")
            # Update case name
            if not sol.step_info.is_solved(self.attempt.cube):
                case_text = sol.step_info.case_name(self.attempt.cube)
                case_label.setText(case_text)
            else:
                case_label.setText("")

        self.attempt.add_cube_listener(refresh)
        cube_widget.layout().addWidget(status_container)
        return cube_widget

    @cached_property
    def current_solution_widget(self) -> "CurrentSolutionWidget":
        w = CurrentSolutionWidget(self.attempt)
        w.installEventFilter(self)
        return w

    @cached_property
    def current_solution_scroll_widget(self):
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(self.current_solution_widget)
        scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        # Force scrollbars to be visible
        scroll_area.setStyleSheet(
            """
            QScrollBar:vertical {
                width: 16px;
                background: #f0f0f0;
            }
            QScrollBar:horizontal {
                height: 16px;
                background: #f0f0f0;
            }
        """
        )

        return scroll_area

    def _empty_container(self, layout) -> QWidget:
        w = QWidget()
        w.setLayout(layout)
        w.setContentsMargins(0, 0, 0, 0)
        w.layout().setContentsMargins(0, 0, 0, 0)
        w.layout().setSpacing(0)
        return w

    @cached_property
    def gui_commands_widget(self) -> QWidget:
        # GUI alternatives to typing commands
        def command_button(cmd: str, name: Optional[str] = None):
            if name is None:
                name = cmd
            b = QPushButton(name)

            @catch_errors
            def execute(*args, **kwargs):
                self.commands.execute(cmd)

            b.clicked.connect(execute)
            return b

        step_selector = QComboBox()
        step_selector.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        step_selector.addItems(
            [
                "",
                "eofb",
                "eorl",
                "eoud",
                "drud",
                "drrl",
                "drfb",
                "htr",
                "fr",
                "finish",
                "insertions",
            ]
        )
        step_selector.setCurrentText("")

        @catch_errors
        def selection_changed(*args, **kwargs):
            text = step_selector.currentText()
            if text:
                self.commands.execute(text)

        step_selector.currentIndexChanged.connect(selection_changed)

        solve_button = QPushButton("solve")
        solve_count = QComboBox()
        solve_count.addItems([str(i) for i in range(1, 21)])

        @catch_errors
        def solve(*args, **kwargs):
            self.commands.execute(f"solve({solve_count.currentIndex()+1})")

        solve_button.clicked.connect(solve)

        gui_widget = self._empty_container(QHBoxLayout())

        w = self._empty_container(QVBoxLayout())
        w.layout().setAlignment(Qt.AlignLeft)
        w.layout().addWidget(step_selector)

        # Preferences button
        pref_button = QPushButton("Display Preferences")

        @catch_errors
        def show(*args, **kwargs):
            prefs.show_dialog(self)

        pref_button.clicked.connect(show)
        w.layout().addWidget(pref_button)

        gui_widget.layout().addWidget(w)

        gui_widget.layout().addWidget(QWidget(), 1)
        gui_widget.layout().addWidget(solve_button)
        gui_widget.layout().addWidget(solve_count)

        w = self._empty_container(QVBoxLayout())
        w.layout().addWidget(command_button("save"))
        w.layout().addWidget(command_button("done"))
        gui_widget.layout().addWidget(w)
        gui_widget.layout().addWidget(QWidget(), 1)

        w = self._empty_container(QVBoxLayout())
        w.layout().addWidget(command_button("niss"))
        h = self._empty_container(QHBoxLayout())
        h.layout().addWidget(command_button("x"))
        h.layout().addWidget(command_button("y"))
        h.layout().addWidget(command_button("z"))
        w.layout().addWidget(h)

        gui_widget.layout().addWidget(w)

        valid_steps: Set[str] = set()

        def enable_items(box: QComboBox, enabled):
            for i in range(box.count()):
                item = box.model().item(i)
                if item.text() in enabled:
                    item.setFlags(item.flags() | Qt.ItemIsEnabled)
                else:
                    item.setFlags(item.flags() & ~Qt.ItemIsEnabled)

        def refresh():
            valid_steps.clear()
            for kind, variant in self.attempt.possible_steps_following("", ""):
                valid_steps.add(f"{kind}{variant}")
            for step in reversed(self.attempt.solution.substeps()):
                valid_steps.add(step_name(step.kind, step.variant))
                if step.step_info.is_solved(self.attempt.cube):
                    for kind, variant in self.attempt.possible_steps_following(
                        step.kind, step.variant
                    ):
                        valid_steps.add(step_name(kind, variant))
            current_step = step_name(
                self.attempt.solution.kind, self.attempt.solution.variant
            )
            enable_items(step_selector, valid_steps)
            step_selector.blockSignals(True)
            step_selector.setCurrentText(current_step)
            step_selector.blockSignals(False)

        self.attempt.add_solution_attribute_listener(refresh)

        return gui_widget

    @cached_property
    def text_input_widget(self) -> QWidget:
        # Text command input
        w = self._empty_container(QHBoxLayout())
        w.layout().setContentsMargins(10, 0, 10, 0)  # Remove all margins
        w.layout().setSpacing(6)

        command_label = QLabel("Command:")
        self.command_input = QLineEdit()
        self.command_input.setText("help")
        self.command_input.setSelection(0, 4)

        @catch_errors
        def execute(*args, **kwargs):
            self.execute_command()

        self.command_input.returnPressed.connect(execute)
        self.command_input.installEventFilter(self)

        w.layout().addWidget(command_label)
        w.layout().addWidget(self.command_input)

        return w

    @cached_property
    def status_widget(self) -> QWidget:
        self.status_label = QLabel()
        self.status_label.setContentsMargins(10, 0, 10, 0)
        return self.status_label

    @cached_property
    def solutions_widget(self) -> QWidget:
        # Lists of solutions for each step
        solutions_container = self._empty_container(QVBoxLayout())
        solutions_container.layout().setContentsMargins(10, 0, 10, 0)  # Remove margins

        solution_lists_layout = QHBoxLayout()
        solution_lists_layout.setSpacing(10)  # Add some spacing between columns

        self.solution_widgets = {}

        # Define colors for different states
        selection_color = "#22dd7c"
        active_selection_color = "#33ff9e"

        # Create a common stylesheet for all list widgets with focus-independent styling
        list_style = f"""
            /* Default appearance */
            QListWidget::item {{
                color: black;
                background-color: transparent;
            }}

            /* Basic selection style (blue) */
            QListWidget::item:selected {{
                background-color: {selection_color};
                color: black;
            }}

            /* Keep selection color even when widget loses focus */
            QListWidget::item:selected:!active {{
                background-color: {active_selection_color};
                color: black;
            }}
        """

        def build_solution_widget(kind: str, label: Optional[str] = None):
            container = self._empty_container(QVBoxLayout())
            label = label or kind.upper()
            container.layout().addWidget(QLabel(f"{label}:"))
            list = QListWidget()
            list.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

            @catch_errors
            def activate(item, *args, **kwargs):
                self.activate_item(item)

            list.itemDoubleClicked.connect(activate)

            @catch_errors
            def selected(*args, **kwargs):
                self.item_selected(list)

            list.itemClicked.connect(selected)
            list.itemSelectionChanged.connect(selected)
            list.installEventFilter(self)
            list.setStyleSheet(list_style)
            list.setItemDelegate(SolutionItemRenderer())
            list.setProperty("kind", kind)
            container.layout().addWidget(list)
            self.solution_widgets[kind] = list
            return container

        # Solution List Widgets
        solution_lists_layout.addWidget(build_solution_widget("eo"))
        solution_lists_layout.addWidget(build_solution_widget("dr"))
        htr_slice = self._empty_container(QVBoxLayout())
        htr_slice.layout().addWidget(build_solution_widget("htr"))
        htr_slice.layout().addWidget(build_solution_widget("finish", "Finish"))
        solution_lists_layout.addWidget(htr_slice)
        fr_finish = self._empty_container(QVBoxLayout())
        fr_finish.layout().addWidget(build_solution_widget("fr"))
        fr_finish.layout().addWidget(build_solution_widget("insertions", "Insertions"))
        solution_lists_layout.addWidget(fr_finish)

        solutions_container.layout().addLayout(solution_lists_layout)

        return solutions_container

    def populate_saved_solutions(self):
        """Clear and refresh the saved solution steps"""
        solutions = self.attempt.solutions_by_kind()

        for kind, list in self.solution_widgets.items():
            list.clear()
            for i, sol in enumerate(solutions.get(kind, [])):
                padding = "   " if i < 9 else ("  " if i < 99 else " ")
                list.addItem(f"{i + 1}.{padding}{self.attempt.to_str(sol)}")
                item = list.item(list.count() - 1)
                item.setData(SOLUTION, sol)
        self.format_saved_solutions()

    def format_saved_solutions(self):
        """
        Apply formatting to the solutions in the solution widgets
        Bold for steps that are active
        Strikethrough for steps marked "done"
        """
        active = self.attempt.solution.substeps()
        active_list_item = None

        for kind in reversed(self.step_order):
            list = self.solution_widgets[kind]
            list.blockSignals(True)
            matched = False
            for i in range(list.count() - 1, -1, -1):
                item = list.item(i)
                sol = item.data(SOLUTION)
                item.setData(BOLD, sol in active and not matched)
                padding = "   " if i < 9 else ("  " if i < 99 else " ")
                item.setText(f"{i + 1}.{padding}{self.attempt.to_str(sol)}")
                matched = matched or item.data(BOLD)
                item.setData(STRIKETHROUGH, self.attempt.is_done(sol))
                item.setSelected(False)
                if item.data(BOLD) and not active_list_item:
                    active_list_item = list, item
            list.blockSignals(False)
            list.update()
        if active_list_item:
            list, item = active_list_item
            item.setSelected(True)
            list.setCurrentItem(item)
            self.item_selected(list)

    def set_status(self, status: str):
        """Show user-relevant info"""
        self.status_label.setText(status)

    def _append_moves(self, moves_str):
        """Append moves to the current solution"""
        alg = Algorithm(moves_str)

        sol = self.attempt.solution

        if sol.alg.len() > 0:
            if not self.attempt.append(alg):
                self.set_status(
                    f"{alg} not allowed after {sol.previous.kind}{sol.previous.variant}"
                )
        else:
            if not self.attempt.append(alg):
                assert sol.previous is not None
                if sol.previous.allows_moves(alg):
                    previous_alg = sol.previous.alg
                    self.attempt.back()
                    was_inverse = self.attempt.inverse
                    self.attempt.set_inverse(False)
                    self.attempt.append(previous_alg)
                    self.attempt.append(alg)
                    self.attempt.set_inverse(was_inverse)
                else:
                    self.set_status(
                        f"{moves_str} not allowed after {sol.previous.kind}{sol.previous.variant}"
                    )

    def set_step(self, kind, variant) -> bool:
        """Change to a specific solving step"""
        step_info = StepInfo(kind, variant)
        sol = self.attempt.solution
        while sol.alg.len() == 0 and sol.previous:
            sol = sol.previous
        past_step_kinds = {s.kind for s in sol.substeps()[:-1]}
        if kind in past_step_kinds:
            # Moving backward
            while sol.kind != kind:
                sol = sol.previous
            if sol.previous:
                sol = sol.previous
            else:
                sol = PartialSolution()
            self.attempt.solution = sol
            self.attempt.advance_to(kind, variant)
            return True
        elif kind == sol.kind:
            self.attempt.back()
            self.attempt.advance_to(kind, variant)
            return True
        else:
            if step_info.is_eligible(self.attempt.cube):
                inverse = self.attempt.inverse
                self.attempt.advance_to(kind, variant)
                self.attempt.set_inverse(inverse)
                return True
            else:
                self.set_status(f"Cube is not eligible for {kind}{variant}")
                return False

    def scroll_to(self, solution: PartialSolution):
        """Make sure the given solution is visible"""
        w = self.solution_widgets[solution.kind]
        for i in range(0, w.count()):
            if w.item(i).data(SOLUTION) == solution:
                w.item(i).setSelected(True)
                w.setCurrentItem(w.item(i))
                w.scrollToItem(w.item(i))

    def activate_item(self, item):
        """
        User has selected a step from one of the solution widgets
        Result of a double-click or hitting enter
        Load the solution step into the view
        """
        sol = item.data(SOLUTION)
        index = self.attempt.solutions_by_kind()[sol.kind].index(sol) + 1
        # Execute via self.commands to get this into the history
        self.commands.execute(f'check("{sol.kind}",{index})')

    def item_selected(self, list_widget):
        """
        User has highlighted a step from one of the solution widgets
        Result of a single-click or arrow key navigation
        Does not activate the step; only shows related steps in the other widgets
        """
        selected_item = list_widget.currentItem()
        if not selected_item:
            return

        selected_step = selected_item.data(SOLUTION)

        for k, w in self.solution_widgets.items():
            w.blockSignals(True)
            w.clearSelection()
            w.setCurrentItem(None)
            w.setSelectionMode(QListWidget.MultiSelection)
            for i in range(w.count()):
                item = w.item(i)
                sol = item.data(SOLUTION)
                highlight = (
                    sol in selected_step.substeps() or selected_step in sol.substeps()
                )
                item.setSelected(highlight)
                if highlight:
                    w.setCurrentItem(item)
            w.setSelectionMode(QListWidget.SingleSelection)
            w.blockSignals(False)
        for _, w in self.solution_widgets.items():
            if w != list_widget:
                w.scrollToItem(w.currentItem())

    def execute_command(self):
        cmd = self.command_input.text().strip()
        self.command_history.append(cmd)
        self.history_pointer = len(self.command_history) - 1
        self.commands.execute(cmd)
        self.command_input.clear()

    def eventFilter(self, obj, event):
        """Handle keyboard events for navigating between solution lists"""
        delegate = super()

        @catch_and_return(False)
        def handle(*args, **kwargs):
            if self.viz.handle_toggle_view_event(obj, event):
                return True
            # Check if this is a key event for one of our solution lists
            if event.type() == QEvent.KeyPress:
                key = event.key()
                if event.isAutoRepeat():
                    return False
                if key == Qt.Key_Return or key == Qt.Key_Enter:
                    if obj in list(self.solution_widgets.values()):
                        kind = obj.property("kind")
                        if not kind:
                            return False
                        # Enter key to check the currently selected solution
                        if obj.currentItem():
                            self.activate_item(obj.currentItem())
                            return True
                    return False
                elif key == Qt.Key_Tab or key == Qt.Key_Backtab:
                    # Handle Tab and Shift+Tab to move between solution lists
                    if (
                        obj not in self.solution_widgets.values()
                        and obj != self.command_input
                    ):
                        return False

                    def select_widget(widget):
                        if widget.count():
                            widget.clearSelection()
                            widget.setCurrentItem(None)
                            selection = 0
                            for i in range(0, widget.count()):
                                if (
                                    widget.item(i).data(SOLUTION)
                                    in self.attempt.solution.substeps()
                                ):
                                    selection = i
                                    break
                            widget.setCurrentItem(widget.item(selection))
                            widget.setFocus()
                            widget.update()
                            return True
                        else:
                            self.command_input.setFocus()
                            return True

                    if obj == self.command_input:
                        # Shift-tab navigates to current_solution_widget
                        if key == Qt.Key_Backtab:
                            self.current_solution_widget.setCurrentItem(
                                self.current_solution_widget.item(
                                    self.current_solution_widget.count() - 1
                                )
                            )
                            self.current_solution_widget.setFocus()
                            return True
                        # Otherwise navigate to the last active step
                        for k in reversed(self.step_order):
                            w = self.solution_widgets[k]
                            for i in range(0, w.count()):
                                if (
                                    w.item(i).data(SOLUTION)
                                    in self.attempt.solution.substeps()
                                ):
                                    return select_widget(w)
                        return select_widget(self.solution_widgets["eo"])
                    # Navigate between solution steps
                    index = self.step_order.index(obj.property("kind"))
                    if key == Qt.Key_Backtab:
                        if index == 0:
                            self.command_input.setFocus()
                            return True
                        else:
                            next_index = index
                            while True:
                                next_index = (next_index - 1) % len(self.step_order)
                                widget = self.solution_widgets[
                                    self.step_order[next_index]
                                ]
                                if next_index == 0 or widget.count() > 0:
                                    break
                            return select_widget(widget)
                    else:
                        if index == len(self.step_order) - 1:
                            self.command_input.setFocus()
                            return True
                        else:
                            next_index = index
                            while True:
                                next_index = (next_index + 1) % len(self.step_order)
                                widget = self.solution_widgets[
                                    self.step_order[next_index]
                                ]
                                if (
                                    next_index == len(self.step_order) - 1
                                    or widget.count() > 0
                                ):
                                    break
                            return select_widget(widget)
                # Command history
                elif obj == self.command_input and key == Qt.Key_Up:
                    if self.history_pointer < 0:
                        return True
                    previous_command = self.command_history[self.history_pointer]
                    self.command_input.setText(previous_command)
                    self.history_pointer = max(0, self.history_pointer - 1)
                    return True
                elif obj == self.command_input and key == Qt.Key_Down:
                    self.history_pointer = min(
                        self.history_pointer + 1, len(self.command_history)
                    )
                    previous_command = ""
                    if self.history_pointer < len(self.command_history) - 1:
                        previous_command = self.command_history[
                            self.history_pointer + 1
                        ]
                    self.command_input.setText(previous_command)
                    self.history_pointer = min(
                        self.history_pointer, len(self.command_history) - 1
                    )
                    return True

            return delegate.eventFilter(obj, event)

        return handle()

    def check_solution(self, solution):
        """Load a selected solution"""
        self.attempt.set_solution(solution)
        if solution.step_info.is_solved(self.attempt.cube):
            self.attempt.advance()
        self.command_input.setFocus()

    def _create_menus(self):
        """Create the menu bar with File and Help menus"""
        # Create menu bar
        menu_bar = self.menuBar()

        # File menu
        file_menu = menu_bar.addMenu("File")

        # Save session action
        save_action = QAction("Save Session", self)

        @catch_errors
        def save(*args, **kwargs):
            self.save_session_dialog()

        save_action.triggered.connect(save)
        file_menu.addAction(save_action)

        # Load session action
        load_action = QAction("Load Session", self)

        @catch_errors
        def load(*args, **kwargs):
            self.load_session_dialog()

        load_action.triggered.connect(load)
        file_menu.addAction(load_action)

        # Export image
        export_action = QAction("Export image", self)

        @catch_errors
        def export(*args, **kwargs):
            self.export_image_dialog()

        export_action.triggered.connect(export)
        file_menu.addAction(export_action)

        # Exit action
        exit_action = QAction("Exit", self)

        @catch_errors
        def close(*args, **kwargs):
            self.close()

        exit_action.triggered.connect(close)
        file_menu.addAction(exit_action)

        # Help menu
        help_menu = menu_bar.addMenu("Help")

        # Help action
        help_action = QAction("Help", self)

        @catch_errors
        def show_help(*args, **kwargs):
            self.show_help()

        help_action.triggered.connect(show_help)
        help_menu.addAction(help_action)

        # About action
        about_action = QAction("About", self)

        @catch_errors
        def show_about(*args, **kwargs):
            self.show_about()

        about_action.triggered.connect(show_about)
        help_menu.addAction(about_action)

    def save_session_dialog(self):
        """Show dialog to save session"""
        filename, _ = QFileDialog.getSaveFileName(
            self, "Save Session", "", "VFMC Files (*.vfmc)"
        )
        if filename:
            if not filename.endswith(".vfmc"):
                filename += ".vfmc"
            self.commands.save_session(filename)

    def load_session_dialog(self):
        """Show dialog to load session"""
        filename, _ = QFileDialog.getOpenFileName(
            self, "Load Session", "", "VFMC Files (*.vfmc)"
        )
        if filename:
            self.commands.load_session(filename)

    def export_image_dialog(self):
        """Show dialog to save session"""
        filename, _ = QFileDialog.getSaveFileName(
            self, "Export Image", "", "PNG (*.png)"
        )
        if filename:
            if not filename.endswith(".png"):
                filename += ".png"
            size, ok = QInputDialog.getInt(
                self, "Image Size", "Enter size:", 100, 25, 500
            )
            if ok:
                self.commands.export_image(filename, size)

    def show_about(self):
        """Show about dialog"""
        about_dialog = QMessageBox(self)
        about_dialog.setWindowTitle("About VFMC")
        about_dialog.setText(f"<h2>VFMC v{version('vfmc')}</h2>")
        about_dialog.setInformativeText(
            "<html><body>"
            "<p style='font-size: 14px;'>A tool for building fewest-move cube solutions virtually.</p>"
            "<p style='font-size: 14px;'>Developed by Rodney Kinney. Based on the cubelib library by Jonas Balsfulland.</p>"
            f"<p style='font-size: 10px;'>Error log: {os.path.join(prefs.app_dir(), 'vfmc.log')}</p>"
            "</body></html>"
        )
        about_dialog.setMinimumWidth(300)
        about_dialog.setMinimumHeight(300)
        about_dialog.setStandardButtons(QMessageBox.Ok)
        about_dialog.show()

    def show_help(self):
        """Show help window"""
        help_dialog = QMessageBox(self)
        help_dialog.setWindowModality(Qt.NonModal)
        help_dialog.setWindowTitle("VFMC help")

        commands = []
        for name in dir(self.commands):
            if name.startswith("_"):
                continue
            attr = getattr(self.commands, name)
            if callable(attr):
                commands.append((name, attr))

        help_file = os.path.join(os.path.dirname(__file__), "help.html")
        with open(help_file, "r") as f:
            help_dialog.setInformativeText(f.read())

        help_dialog.setText("Welcome to VFMC")
        help_dialog.setStandardButtons(QMessageBox.Ok)
        help_dialog.show()
        self.command_input.setFocus()


class InsertionsDialog(QDialog):
    def __init__(self, window, step):
        super().__init__(parent=window)
        self.window = window
        self.step = step
        self.setWindowTitle("Insertions")
        self.setModal(True)
        self.resize(800, 100)

        layout = QVBoxLayout(self)
        self.attempt = Attempt()
        self.attempt.set_scramble(self.window.attempt.scramble)
        self.attempt.set_solution(self.step)
        viz = CubeViz(self.attempt)
        layout.addWidget(CubeWidget(viz))

        self.save_button = QPushButton("Save", self)

        @catch_errors
        def save(*args, **kwargs):
            self.save()

        self.save_button.clicked.connect(save)
        self.save_button.setEnabled(False)

        # Editable text field
        self.text_edit = QLineEdit(self)

        @catch_errors
        def refresh(*args, **kwargs):
            self.refresh()

        self.text_edit.cursorPositionChanged.connect(refresh)
        self.text_edit.textChanged.connect(refresh)
        self.text_edit.setText(
            str(Algorithm("").merge(self.step.insertions.replacement.all_on_normal()))
        )
        layout.addWidget(self.text_edit)

        # Buttons
        buttons = QHBoxLayout()
        layout.addLayout(buttons)
        buttons.addStretch(1)
        buttons.addWidget(self.save_button)
        cancel_button = QPushButton("Cancel", self)

        @catch_errors
        def close(*args, **kwargs):
            self.close()

        cancel_button.clicked.connect(close)
        buttons.addWidget(cancel_button)

        self.setLayout(layout)

    def refresh(self):
        self.step.set_replacement(
            self.text_edit.text(), self.text_edit.cursorPosition()
        )
        self.attempt.set_solution(self.step)
        step = self.step.clone()
        step.set_replacement(self.text_edit.text(), len(self.text_edit.text()))
        is_solved = step.step_info.is_solved(
            Cube(f"{self.attempt.scramble} {step.full_alg()}")
        )
        self.save_button.setEnabled(is_solved)

    def save(self):
        self.window.commands.execute(f'compute_insertions("{self.text_edit.text()}")')
        self.window.commands.execute('comment("solved")')
        self.window.commands.execute("save")
        self.close()


def run(session_file: Optional[str] = None):
    app = QApplication(sys.argv)
    app.setApplicationName("VFMC")
    window = AppWindow()
    window.show()
    if session_file:
        window.commands.load_session(session_file)

    sys.exit(app.exec_())


class Commands:
    def __init__(self, window: AppWindow):
        self.window = window
        self.attempt = window.attempt
        self.command_history = []  # Commands that were executed
        self.history_pointer = -1

    def execute(self, raw_command):
        """Execute a command from the command input"""
        cmd = raw_command
        if not cmd:
            return

        try:
            self.window.set_status("")
            result = None
            # Does it match a method in this class?
            if cmd.endswith("'"):
                cmd = cmd.replace("'", "_prime")
            cmd_name = re.sub("\\(.*$", "", cmd)
            if cmd_name in dir(self):
                # Add parentheses if missing
                if cmd.find("(") < 0:
                    cmd = f"{cmd}()"
                # Use locals and globals from this context
                local_vars = {"self": self}
                exec(f"result = self.{cmd}", globals(), local_vars)
                result = local_vars.get("result")
                if result is None:
                    result = CommandResult(add_to_history=[raw_command])
                if result.error is not None:
                    self.window.set_status(f"Error: {result.error}")
                elif result.add_to_history:
                    self.command_history.extend(result.add_to_history)
                    self.history_pointer = len(self.command_history) - 1
            else:
                # Check if it's a sequence of cube moves
                if re.fullmatch(MOVE_REGEX, raw_command):
                    self.window._append_moves(raw_command)
                    self.command_history.append(raw_command)
                else:
                    self.window.set_status(f"No such command: {raw_command}")
        except Exception as e:
            logging.error(traceback.format_exc())
            logging.error(sys.exc_info())
            self.window.set_status(f"Error: {e}")

    def debug(self):
        from vfmc_core import debug

        self.window.set_status(debug(self.attempt.cube))
        return CommandResult(add_to_history=[])

    def help(self):
        self.window.show_help()
        return CommandResult(add_to_history=[])

    def x(self):
        self.attempt.solution.x(1)

    def x_prime(self):
        self.attempt.solution.x(3)

    def x2(self):
        self.attempt.solution.x(2)

    def y(self):
        self.attempt.solution.y(1)

    def y_prime(self):
        self.attempt.solution.y(3)

    def y2(self):
        self.attempt.solution.y(2)

    def z(self):
        self.attempt.solution.z(1)

    def z_prime(self):
        self.attempt.solution.z(3)

    def z2(self):
        self.attempt.solution.z(2)

    def set_step(self, kind, variant):
        self.window.set_step(kind, variant)

    def eoud(self):
        self.window.set_step("eo", "ud")

    def eofb(self):
        self.window.set_step("eo", "fb")

    def eorl(self):
        self.window.set_step("eo", "rl")

    def drud(self):
        self.window.set_step("dr", "ud")

    def drfb(self):
        self.window.set_step("dr", "fb")

    def drrl(self):
        self.window.set_step("dr", "rl")

    def htr(self):
        variant = ""
        for v in ["ud", "fb", "rl"]:
            if StepInfo("dr", v).is_solved(self.window.attempt.cube):
                variant = v
                break
        if variant:
            self.window.set_step("htr", variant)
        else:
            self.window.set_status("Cube is not eligible for HTR")

    def fr(self, axis=None):
        """Look for FR"""
        variant = axis
        if variant is None:
            variant = next(
                s.variant for s in self.attempt.solution.substeps() if s.kind == "dr"
            )
        if variant is not None:
            self.window.set_step("fr", variant)
        else:
            self.window.set_status(
                """No DR step found. Specify axis="..." to set the FR axis"""
            )

    def finish(self, axis=None):
        self.window.set_step("finish", "")

    def insertions(self):
        self.window.set_step("insertions", "")

    def compute_insertions(self, moves_str: str):
        if self.attempt.solution.kind != "insertions":
            return CommandResult(add_to_history=[])
        self.attempt.solution.set_replacement(moves_str, len(moves_str))
        self.attempt.solution.add_markers()

    def set_inverse(self, b: bool):
        self.attempt.set_inverse(b)

    def niss(self):
        """Switch between normal and inverse scramble"""
        self.attempt.set_inverse(not self.attempt.inverse)

    def solve(self, num_solutions: int = 1):
        """Find and save solutions for the current step"""
        if num_solutions > 50:
            self.window.set_status("Maximum of 50 solutions per solve")
            return
        self.window.set_status(
            f"Finding solutions for {self.attempt.solution.kind}{self.attempt.solution.variant}..."
        )
        solutions = self.attempt.solve(num_solutions)
        if solutions:
            self.window.set_status(
                f"Found {len(solutions)} solution{'' if len(solutions) == 1 else 's'}"
            )
        else:
            self.window.set_status(
                f"No solutions found for {self.attempt.solution.kind}{self.attempt.solution.variant}"
            )
        commands = []
        if solutions:
            was_inverse = self.attempt.inverse
            for sol in solutions:
                commands.append(step_name(sol.kind, sol.variant))
                if sol.alg.normal_moves():
                    commands.append("set_inverse(False)")
                    commands.append(" ".join(sol.alg.normal_moves()))
                if sol.alg.inverse_moves():
                    commands.append("set_inverse(True)")
                    commands.append(" ".join(sol.alg.inverse_moves()))
                commands.append("save")
            commands.append(f"set_inverse({was_inverse})")
            commands.append(
                step_name(self.attempt.solution.kind, self.attempt.solution.variant)
            )

        for cmd in commands:
            self.execute(cmd)
        if len(solutions) == 1:
            sol = solutions[0]
            index = self.attempt.solutions_by_kind()[sol.kind].index(sol) + 1
            self.execute(f'check("{sol.kind}",{index})')
        return CommandResult(add_to_history=[])

    def comment(self, s: str):
        sol = self.attempt.solution
        if sol.alg.len() == 0:
            sol = sol.previous
        if not sol:
            return
        self.attempt.set_comment(sol, s)
        item = self._item_containing(sol)
        if item:
            i = item.listWidget().row(item)
            padding = "   " if i < 9 else ("  " if i < 99 else " ")
            item.setText(f"{i + 1}.{padding}{self.attempt.to_str(sol)}")
            return

    def _item_containing(self, sol: PartialSolution) -> Optional[QListWidgetItem]:
        for kind in reversed(self.window.step_order):
            widget = self.window.solution_widgets[kind]
            for i in range(widget.count()):
                item = widget.item(i)
                if item.data(SOLUTION) == sol:
                    return item
        return None

    def done(self):
        sol = self.attempt.solution
        self.attempt.toggle_done(sol)

    def obscure(self):
        sol = self.attempt.solution
        self.attempt.toggle_obscured(sol)

    def save(self):
        """Save this algorithm and start a new one"""
        self.window.current_solution_widget.sync_history_with_editor()
        saved = self.attempt.save()
        if saved:
            self.window.scroll_to(saved)
            self.window.command_input.setFocus()

    def reset(self):
        """Reset the cube to the beginning of the current step"""
        self.attempt.reset()

    def back(self):
        """Go back to the previous step"""
        self.attempt.back()

    def check(self, kind: str, index: int):
        k = kind.lower()
        if k not in self.attempt.solutions_by_kind():
            self.window.set_status(f"Bad step type: {kind}")
            return
        solutions = self.attempt.solutions_by_kind()[kind.lower()]
        if index < 1 or index > len(solutions):
            self.window.set_status(f"Couldn't find {kind} #{index}")
            return
        self.window.check_solution(solutions[index - 1].clone())

    def scramble(self, scramble: str = None):
        """
        <br>Use scramble(\"...\") to initialize with the specified scramble
        <br>or omit the parentheses to generate a new random scramble
        """
        if scramble is None:
            wrapper = Algorithm("R' U' F")
            scramble = str(wrapper.merge(Algorithm(gen_scramble())).merge(wrapper))
        self.attempt.set_scramble(scramble)
        self.command_history.clear()
        return CommandResult(add_to_history=[f'scramble("{scramble}")'])

    def save_session(self, filename):
        try:
            with open(filename, "w") as f:
                f.write(f"# VFMC version {version('vfmc')}\n")
                f.writelines("\n".join(self.command_history))
            self.window.set_status(f"Saved session to {filename}")
            return CommandResult(add_to_history=[])
        except Exception as e:
            self.window.set_status(f"Unable to save to {filename}: {e}")
            return CommandResult(error=e, add_to_history=[])

    def load_session(self, filename):
        h = self.command_history
        try:
            self.command_history = []
            with open(filename, "r") as f:
                for cmd in f.readlines():
                    if not cmd.startswith("#"):
                        self.execute(cmd.strip())
            self.window.set_status(f"Loaded '{filename}'")
            return CommandResult(add_to_history=[])
        except Exception as e:
            self.command_history = h
            self.window.set_status(f"Unable to load '{filename}': {e}")
            return CommandResult(error=e, add_to_history=[])

    def palette(self, name):
        p = Palette.by_name(name) if name else None
        self.window.viz.set_palette(p)
        return CommandResult(add_to_history=[])

    def camera(self, x, y, z, angle=-math.pi / 6):
        self.window.viz.set_camera(x, y, z)
        self.window.viz.view_y = angle
        self.window.viz.refresh()
        return CommandResult(add_to_history=[])

    def export_image(self, filename, size=None):
        self.window.cube_viz_widget.export_png(filename, size)
        return CommandResult(add_to_history=[])


class SolutionItemRenderer(QStyledItemDelegate):
    def __init__(self):
        super(QStyledItemDelegate, self).__init__()

    def initStyleOption(self, option, item):
        # Override style for the active solution
        super().initStyleOption(option, item)

        if item.data(STRIKETHROUGH):
            option.font.setStrikeOut(True)
        if item.data(BOLD):
            option.font.setBold(True)


class CurrentSolutionWidget(QListWidget):
    def __init__(self, attempt: Attempt):
        super().__init__()
        self.attempt = attempt
        self.history_is_stale = False
        self.current_editor = None
        self.comment = None
        self.originalKeyPress = None
        self.original_alg = None
        self.setSelectionMode(QListWidget.ContiguousSelection)
        self.setStyleSheet("font-size: 16px;")
        self.setEditTriggers(QListWidget.DoubleClicked)
        self.attempt.add_cube_listener(self.sync_widget_with_cube)

    def sync_widget_with_cube(self):
        """Sync the widget with Attempt cube state"""
        if self.current_editor:
            return  # Don't sync if user is editing
        self.clear()
        self.addItem(self.attempt.scramble)
        self.addItem("")
        for step in self.attempt.solution.substeps():
            text = f"{step.alg_str(verbose=True)}"
            if step.kind != "":
                if step.step_info.is_solved(self.attempt.cube):
                    text = f"{self.attempt.to_str(step, verbose=True)}"
                else:
                    parentheses = f"{' ' if step.alg.len() else ''}{'( )' if not step.alg.inverse_moves() and self.attempt.inverse else ''}"
                    comment = self.attempt.get_comment(step)
                    if not comment or step.alg.len() == 0:
                        comment = step.default_comment()
                    text = f"{text}{parentheses} // {comment} ({step.full_alg().len()})"
            item = QListWidgetItem(text)
            item.setData(SOLUTION, step)
            self.addItem(item)
        last_item = self.item(self.count() - 1)
        if last_item.data(SOLUTION).kind != "insertions":
            last_item.setFlags(last_item.flags() | Qt.ItemIsEditable)

    def edit(self, index, trigger, event):
        delegate = super()

        @catch_and_return(False)
        def handle(*args, **kwargs):
            result = delegate.edit(index, trigger, event)
            if result:
                editor = self.indexWidget(index)
                if editor:
                    self.current_editor = editor
                    text = editor.text().split("//")[0].strip()
                    editor.setText(text)
                    if ")" in text and self.attempt.inverse:
                        editor.setCursorPosition(text.index(")"))
                    self.original_alg = self.attempt.solution.alg

                    @catch_errors
                    def sync(*args, **kwargs):
                        self.sync_cube_with_editor()

                    self.current_editor.textEdited.connect(sync)
                    self.comment = None
            return result

        return handle()

    def sync_cube_with_editor(self):
        if not self.current_editor:
            return
        edited_text = self.current_editor.text().split("//")
        alg_str = edited_text[0].strip()
        self.comment = edited_text[1].strip() if len(edited_text) > 1 else None
        if "(" in alg_str and ")" not in alg_str:
            return
        if not alg_str and self.attempt.inverse:
            self.current_editor.setText("( )")
            self.current_editor.setCursorPosition(2)
        try:
            alg = Algorithm(alg_str)
            self.attempt.solution.alg = Algorithm("")
            if not self.attempt.solution.append(alg):
                self.window().set_status(
                    f"{alg} not allowed after {self.attempt.solution.previous.kind}{self.attempt.solution.previous.variant}"
                )
            else:
                self.window().set_status("")
                self.attempt.update_cube()
                self.history_is_stale = True
        except Exception:
            pass

    def sync_history_with_editor(self):
        if not self.history_is_stale or not self.original_alg:
            return
        # Reset solution to original state, before we started editing
        current_alg = self.attempt.solution.alg
        self.attempt.solution.alg = self.original_alg

        # To properly populate the history, execute the moves done in the editor
        net_alg = self.original_alg.inverted().merge(current_alg)
        commands = self.parent().window().commands
        if net_alg.normal_moves():
            if self.attempt.inverse:
                commands.execute("niss")
            commands.execute(" ".join(net_alg.normal_moves()))
        if net_alg.inverse_moves():
            if not self.attempt.inverse:
                commands.execute("niss")
            commands.execute(" ".join(net_alg.inverse_moves()))
        if self.comment:
            commands.execute(f'comment("{self.comment}")')
        self.history_is_stale = False

    def closeEditor(self, editor, hint):
        # Called when editing is finished
        delegate = super()

        @catch_errors
        def handle(*args, **kwargs):
            self.sync_history_with_editor()
            self.parent().window().command_input.setFocus()

            self.current_editor = None
            self.original_alg = None

            delegate.closeEditor(editor, hint)
            self.sync_widget_with_cube()

        handle()

    def mouseDoubleClickEvent(self, event):
        delegate = super()

        @catch_errors
        def handle(*args, **kwargs):
            item = self.itemAt(event.pos())
            if item:
                self.activate_step(item.data(SOLUTION))
            delegate.mouseDoubleClickEvent(event)

        handle()

    # Override key event to handle copy and editing
    def keyPressEvent(self, event):
        delegate = super()

        @catch_errors
        def handle(*args, **kwargs):
            if event.matches(QKeySequence.Copy):
                selected_items = self.selectedItems()
                if selected_items:
                    clipboard_text = "\n".join(item.text() for item in selected_items)
                    QApplication.clipboard().setText(clipboard_text)
            elif event.key() in {
                Qt.Key_Enter,
                Qt.Key_Return,
                Qt.Key_Delete,
                Qt.Key_Backspace,
            }:
                # Enter key starts editing the current item if any
                current_item = self.currentItem()
                if current_item:
                    if current_item.flags() & Qt.ItemIsEditable:
                        self.editItem(current_item)
                    else:
                        self.activate_step(current_item.data(SOLUTION))
            else:
                delegate.keyPressEvent(event)  # Default behavior for other keys

        handle()

    def activate_step(self, target: PartialSolution):
        if target.kind == "insertions":
            if not self.attempt.solution.step_info.is_solved(self.attempt.cube):
                InsertionsDialog(self.window(), self.attempt.solution).exec()
            return
        if target not in self.attempt.solutions_by_kind()[target.kind]:
            return
        # Execute via self.commands to get this into the history
        cmd = self.window().commands
        cmd.execute(f'set_step("{target.kind}","{target.variant}")')
        inverse = self.attempt.inverse
        if target.alg.inverse_moves():
            cmd.execute("set_inverse(True)")
            cmd.execute(" ".join(target.alg.inverse_moves()))
        if target.alg.normal_moves():
            cmd.execute("set_inverse(False)")
            cmd.execute(" ".join(target.alg.normal_moves()))
        cmd.execute(f"set_inverse({inverse})")
        self.editItem(self.item(self.count() - 1))
        self.window().format_saved_solutions()


@dataclasses.dataclass
class CommandResult:
    error: Optional[Exception] = None
    add_to_history: List[str] = None


def main():
    # Configure logging
    logfile = prefs.app_dir() / "vfmc.log"
    logging.basicConfig(
        filename=logfile,
        filemode="w",
        level=logging.DEBUG,
        format="%(levelname)s - %(message)s",
    )

    if getattr(sys, "frozen", False):  # Running from a PyInstaller bundle
        bundle_dir = os.path.dirname(sys.executable)
        # Set working directory to the home of the executable
        if os.path.basename(bundle_dir) == "MacOS":
            logging.debug(f"Running bundle from {bundle_dir}")
            os.chdir(os.path.dirname(os.path.dirname(bundle_dir)))
    session_file = os.path.abspath(sys.argv[1]) if len(sys.argv) > 1 else None
    run(session_file)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import List
from difflib import SequenceMatcher
import re

from vfmc.orientation import AXIS_ROTATIONS, OPPOSITES
from vfmc_core import Algorithm


@dataclass
class Insertion:
    pos: int
    moves: List[str]


@dataclass
class Replacement:
    start: int
    end: int
    moves: List[str]


class Insertions:
    def __init__(self, alg: Algorithm):
        self.original = alg
        self.replacement = alg
        self.wide_alg_moves = alg.all_on_normal().normal_moves()

    def set_replacement(self, moves_str: str, pos: int):
        try:
            assert "(" not in moves_str and ")" not in moves_str
            before_cursor = parse_wide_alg(moves_str[:pos])
            after_cursor = parse_wide_alg(moves_str[pos:])
            alg = Algorithm(
                f"{' '.join(before_cursor.normalized_moves)} ({Algorithm(' '.join(after_cursor.normalized_moves)).inverted()})"
            )
            self.replacement = alg
            self.wide_alg_moves = before_cursor.wide_moves
        except Exception:
            pass

    def net_alg(self) -> Algorithm:
        return self.original.inverted().merge(self.replacement)

    def get_edits(self) -> List:
        old_moves = self.original.all_on_normal().normal_moves()
        new_moves = self.wide_alg_moves
        m = SequenceMatcher(a=old_moves, b=new_moves)
        edits = []
        for tag, i1, i2, j1, j2 in m.get_opcodes():
            if tag == "equal":
                pass
            elif tag == "replace":
                edits.append(Replacement(start=i1, end=i2, moves=new_moves[j1:j2]))
                pass
            elif tag == "delete":
                edits.append(
                    Insertion(
                        pos=i1,
                        moves=Algorithm(" ".join(old_moves[i1:i2]))
                        .inverted()
                        .normal_moves(),
                    )
                )
            elif tag == "insert":
                edits.append(Insertion(pos=i1, moves=new_moves[j1:j2]))
            else:
                raise f"Unknown tag: {tag}"
        return edits


@dataclass
class WideAlgorithm:
    wide_moves: List[str]
    normalized_moves: List[str]


def parse_wide_alg(moves_str) -> WideAlgorithm:
    """Parse an algorithm that accepts wide moves"""
    moves_str = moves_str.lower()
    alg_pattern = r"^\s*([rufldb][w]?[2']?\s*)*$"
    if not re.fullmatch(alg_pattern, moves_str):
        raise ValueError("Invalid algorithm")
    move_pattern = r"([rufldb])([w]?)([2']?)(?:\s*)"
    moves = re.findall(move_pattern, moves_str)

    def create_transform(f, r):
        rot = AXIS_ROTATIONS[f]
        ticks = 3 if r == "'" else (2 if r == "2" else 1)

        def transform(f):
            if f in rot:
                ft = rot[(rot.index(f) + ticks) % 4]
            else:
                ft = f
            return ft

        return transform

    transformations = []
    wide_moves = []
    norm_moves = []
    for face, wide, rotation in moves:
        wide_moves.append(f"{face.upper()}{wide}{rotation}")
        is_wide = wide == "w"
        for t in transformations:
            face = t(face)

        if is_wide:
            transformations.append(create_transform(face, rotation))
            face = OPPOSITES[face]

        norm_moves.append(f"{face}{rotation}")

    return WideAlgorithm(wide_moves=wide_moves, normalized_moves=norm_moves)



# Rotation matrix for different faces
AXIS_ROTATIONS = {
    "f": ["u", "l", "d", "r"],
    "b": ["u", "r", "d", "l"],
    "r": ["u", "f", "d", "b"],
    "l": ["u", "b", "d", "f"],
    "u": ["f", "r", "b", "l"],
    "d": ["f", "l", "b", "r"],
}

OPPOSITES = {
    "f": "b",
    "b": "f",
    "r": "l",
    "l": "r",
    "u": "d",
    "d": "u",
}


# Default cube orientations for different steps
VARIANT_ORIENTATIONS = {
    "": {"": ("u", "f")},
    "eo": {
        "ud": ("b", "u"),
        "fb": ("u", "f"),
        "rl": ("u", "r"),
    },
    "*": {
        "ud": ("u", "f"),
        "fb": ("b", "u"),
        "rl": ("r", "f"),
        "*": ("u", "f"),
    },
}


class Orientation:
    def __init__(self, top: str, front: str):
        self.top = top
        self.front = front

    def x(self, ticks: int) -> "Orientation":
        rot = AXIS_ROTATIONS[self.right]
        return Orientation(
            top=rot[(rot.index(self.top) + ticks) % 4],
            front=rot[(rot.index(self.front) + ticks) % 4],
        )

    @property
    def right(self):
        rot = AXIS_ROTATIONS[self.top]
        return rot[(rot.index(self.front) + 1) % 4]

    def z(self, ticks: int) -> "Orientation":
        rot = AXIS_ROTATIONS[self.front]
        return Orientation(top=rot[(rot.index(self.top) + ticks) % 4], front=self.front)

    def y(self, ticks: int) -> "Orientation":
        rot = AXIS_ROTATIONS[self.top]
        return Orientation(top=self.top, front=rot[(rot.index(self.front) + ticks) % 4])

    def __repr__(self):
        return f"top={self.top}, front={self.front}"

    @staticmethod
    def default_for(kind, variant) -> "Orientation":
        d = VARIANT_ORIENTATIONS.get(kind, VARIANT_ORIENTATIONS.get("*"))
        return Orientation(*d.get(variant, d.get("*")))



import os
import math
import numpy as np
from PyQt5.QtCore import QTimer, Qt, QEvent, QSize
from PyQt5.QtWidgets import QWidget
from PyQt5.QtGui import QPainter, QColor, QPen, QBrush, QPolygon, QPixmap


from vfmc.attempt import Attempt
from vfmc.orientation import Orientation, AXIS_ROTATIONS
from vfmc.palette import FaceletColors, Visibility, Palette
from vfmc.prefs import preferences
from pyquaternion import Quaternion

# X coordinate of cube facelets
# U + L + F + R + B + D
facelet_x = (
    [-1, 0, 1, -1, 0, 1, -1, 0, 1]
    + [-1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5]
    + [-1, 0, 1, -1, 0, 1, -1, 0, 1]
    + [1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5]
    + [1, 0, -1, 1, 0, -1, 1, 0, -1]
    + [-1, 0, 1, -1, 0, 1, -1, 0, 1]
)

# Y coordinate of cube facelets
# U + L + F + R + B + D
facelet_y = (
    [1, 1, 1, 0, 0, 0, -1, -1, -1]
    + [1, 0, -1, 1, 0, -1, 1, 0, -1]
    + [-1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5]
    + [-1, 0, 1, -1, 0, 1, -1, 0, 1]
    + [1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5]
    + [-1, -1, -1, 0, 0, 0, 1, 1, 1]
)

# Z coordinate of cube facelets
# U + L + F + R + B + D
facelet_z = (
    [1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5]
    + [1, 1, 1, 0, 0, 0, -1, -1, -1]
    + [1, 1, 1, 0, 0, 0, -1, -1, -1]
    + [1, 1, 1, 0, 0, 0, -1, -1, -1]
    + [1, 1, 1, 0, 0, 0, -1, -1, -1]
    + [-1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5, -1.5]
)

# Coordinates of the facelet in different plains, relative to the facelet center
FACELET_WIDTH = float(
    os.getenv("FACELET_WIDTH", "0.48")
)  # Distance from center to edge of the facelet
FACELET_VERTICES = {
    "xy": [
        np.array([-FACELET_WIDTH, -FACELET_WIDTH, 0.0]),
        np.array([FACELET_WIDTH, -FACELET_WIDTH, 0.0]),
        np.array([FACELET_WIDTH, FACELET_WIDTH, 0.0]),
        np.array([-FACELET_WIDTH, FACELET_WIDTH, 0.0]),
    ],
    "xz": [
        np.array([-FACELET_WIDTH, 0.0, -FACELET_WIDTH]),
        np.array([FACELET_WIDTH, 0.0, -FACELET_WIDTH]),
        np.array([FACELET_WIDTH, 0.0, FACELET_WIDTH]),
        np.array([-FACELET_WIDTH, 0.0, FACELET_WIDTH]),
    ],
    "yz": [
        np.array([0.0, -FACELET_WIDTH, -FACELET_WIDTH]),
        np.array([0.0, FACELET_WIDTH, -FACELET_WIDTH]),
        np.array([0.0, FACELET_WIDTH, FACELET_WIDTH]),
        np.array([0.0, -FACELET_WIDTH, FACELET_WIDTH]),
    ],
}

# Plane in which each facelet exists
# U + L + F + R + B + D
FACELET_AXIS = (
    ["xy"] * 9 + ["yz"] * 9 + ["xz"] * 9 + ["yz"] * 9 + ["xz"] * 9 + ["xy"] * 9
)

# Colors of the corner pieces, for orientation 0,1,2
CORNER_PIECE_COLORS = [
    (FaceletColors.WHITE, FaceletColors.ORANGE, FaceletColors.BLUE),
    (FaceletColors.WHITE, FaceletColors.BLUE, FaceletColors.RED),
    (FaceletColors.WHITE, FaceletColors.RED, FaceletColors.GREEN),
    (FaceletColors.WHITE, FaceletColors.GREEN, FaceletColors.ORANGE),
    (FaceletColors.YELLOW, FaceletColors.ORANGE, FaceletColors.GREEN),
    (FaceletColors.YELLOW, FaceletColors.GREEN, FaceletColors.RED),
    (FaceletColors.YELLOW, FaceletColors.RED, FaceletColors.BLUE),
    (FaceletColors.YELLOW, FaceletColors.BLUE, FaceletColors.ORANGE),
]

# Index of the facelet (orientation 0,1,2) for each of the corners
CORNER_POSITION_FACELETS = [
    (0, 9, 38),  # UBL
    (2, 36, 29),  # UBR
    (8, 27, 20),  # UFR
    (6, 18, 11),  # UFL
    (45, 17, 24),  # DFL
    (47, 26, 33),  # DFR
    (53, 35, 42),  # DBR
    (51, 44, 15),  # DBL
]

# Colors of the edge pieces, for orientation 0,1
EDGE_PIECE_COLORS = [
    (FaceletColors.WHITE, FaceletColors.BLUE),
    (FaceletColors.WHITE, FaceletColors.RED),
    (FaceletColors.WHITE, FaceletColors.GREEN),
    (FaceletColors.WHITE, FaceletColors.ORANGE),
    (FaceletColors.GREEN, FaceletColors.RED),
    (FaceletColors.GREEN, FaceletColors.ORANGE),
    (FaceletColors.BLUE, FaceletColors.RED),
    (FaceletColors.BLUE, FaceletColors.ORANGE),
    (FaceletColors.YELLOW, FaceletColors.GREEN),
    (FaceletColors.YELLOW, FaceletColors.RED),
    (FaceletColors.YELLOW, FaceletColors.BLUE),
    (FaceletColors.YELLOW, FaceletColors.ORANGE),
]

# Index of the facelet (orientation 0,1) for each of the edges
EDGE_POSITION_FACELETS = [
    (1, 37),  # UB
    (5, 28),  # UR
    (7, 19),  # UF
    (3, 10),  # UL
    (23, 30),  # FR
    (21, 14),  # FL
    (39, 32),  # BR
    (41, 12),  # BL
    (46, 25),  # DF
    (50, 34),  # DR
    (52, 43),  # DB
    (48, 16),  # DL
]

# The slice where each edge belongs, when the cube is solved
# 0 = M, 1 = E, 2 = S
HOME_SLICE = [0, 2, 0, 2, 1, 1, 1, 1, 0, 2, 0, 2]

# Orientation (in cubelib bit representation) for a correctly-oriented edge
# Index is the combination of the edge's home slice and the slice where it currently is
DEFAULT_ORIENTATION = [
    0,  # Piece is in its home slice
    5,  # E <-> M
    4,  # M <-> S
    1,  # E <-> S
]


class CubeViz:
    """Cube visualization logic"""

    def __init__(
        self,
        attempt: Attempt,
    ):
        self.attempt = attempt
        self.attempt.add_cube_listener(self.refresh)
        preferences.add_listener(self.refresh)

        # Initial camera position
        self.set_camera(0, -10, 6)
        self.view_y = -math.pi / 6
        self.view_x = 0

        self.colors = [(1, 1, 1, 0.2)] * 54
        self.palette = None
        self.hide_nearest_faces = False

    def set_camera(self, x, y, z):
        self.camera = np.array([x, y, z])

        z_axis = np.array([0, 0, 1])
        screen_x_dir = np.cross(-self.camera, z_axis)
        self.screen_x_dir = screen_x_dir / np.linalg.norm(screen_x_dir)
        screen_y_dir = np.cross(self.camera, self.screen_x_dir)
        self.screen_y_dir = screen_y_dir / np.linalg.norm(screen_y_dir)

    def set_palette(self, p: Palette):
        self.palette = p
        self.refresh()

    def get_palette(self) -> Palette:
        return self.palette or Palette.by_name(self.attempt.solution.kind)

    def handle_toggle_view_event(self, obj, event):
        if (
            event.type() == QEvent.KeyPress
            and event.key() == Qt.Key_Alt
            and not event.isAutoRepeat()
        ):
            self.hide_nearest_faces = True
            return True
        if (
            event.type() == QEvent.KeyRelease
            and event.key() == Qt.Key_Alt
            and not event.isAutoRepeat()
        ):
            self.hide_nearest_faces = False
            return True
        return False

    def refresh(self):
        palette = self.get_palette()
        self.colors = [palette.hidden_color] * 54
        self.colors[4] = palette.color_of_center(FaceletColors.WHITE, Visibility.All)
        self.colors[13] = palette.color_of_center(FaceletColors.ORANGE, Visibility.All)
        self.colors[22] = palette.color_of_center(FaceletColors.GREEN, Visibility.All)
        self.colors[31] = palette.color_of_center(FaceletColors.RED, Visibility.All)
        self.colors[40] = palette.color_of_center(FaceletColors.BLUE, Visibility.All)
        self.colors[49] = palette.color_of_center(FaceletColors.YELLOW, Visibility.All)
        corners = self.attempt.cube.corners()
        corner_visibility = self.attempt.corner_visibility()
        for i in range(0, 8):
            piece_id, orientation = corners[i]
            for side in range(0, 3):
                face = (side + 3 - orientation) % 3
                self.colors[CORNER_POSITION_FACELETS[i][side]] = (
                    palette.color_of_corner(
                        CORNER_PIECE_COLORS[piece_id][face], corner_visibility[i][side]
                    )
                )
        edges = self.attempt.cube.edges()
        edge_visibility = self.attempt.edge_visibility()
        for i in range(0, 12):
            piece_id, piece_orientation = edges[i]
            orientation = DEFAULT_ORIENTATION[HOME_SLICE[piece_id] ^ HOME_SLICE[i]]
            flipped = 0 if piece_orientation == orientation else 1
            for side in range(0, 2):
                self.colors[EDGE_POSITION_FACELETS[i][side]] = palette.color_of_edge(
                    EDGE_PIECE_COLORS[edges[i][0]][(side + flipped) % 2],
                    edge_visibility[i][side],
                )

    def draw_facelet(self, painter, w, h, rotation_matrix, x, y, z, color, axis):
        # Define the vertices of the square face based on axis

        vertex_center = np.array([x, y, z])
        # Transform vertices to world coordinates
        transformed_vertices = [
            rotation_matrix @ (v + vertex_center) for v in FACELET_VERTICES[axis]
        ]

        scale_factor = min(w, h) * np.linalg.norm(self.camera) / 5
        screen_vertices = [
            (
                w / 2
                + scale_factor
                * np.dot(v, self.screen_x_dir)
                / np.linalg.norm(v - self.camera),
                h / 2
                - scale_factor
                * np.dot(v, self.screen_y_dir)
                / np.linalg.norm(v - self.camera),
            )
            for v in transformed_vertices
        ]

        # Enable antialiasing for smoother edges
        painter.setRenderHint(QPainter.Antialiasing)

        # Create a semi-transparent pen with alpha channel (RGBA)
        qcol = QColor(*color)
        pen = QPen(QColor(255, 255, 255, 16), 1)
        painter.setPen(pen)

        # Create a semi-transparent brush with alpha channel
        brush = QBrush(qcol)
        painter.setBrush(brush)
        from PyQt5 import QtCore

        polygon = QPolygon([QtCore.QPoint(*v) for v in screen_vertices])
        painter.drawPolygon(polygon)

    def set_inverse(self, inverse: bool):
        self.inverse = inverse

    def update(self):
        self.refresh()

    def draw(self, painter, w, h, fill_bg=True):
        if fill_bg:
            # Clear the screen with the background color
            bg = preferences.background_color
            painter.fillRect(0, 0, w, h, QColor(bg, bg, bg))
        # Apply rotation
        q = (
            Quaternion(axis=[1, 0, 0], angle=self.view_x)
            * Quaternion(axis=[0, 0, 1], angle=self.view_y)
            * rotation_for(self.attempt.solution.orientation)
        )
        rotation_matrix = q.rotation_matrix

        # Order faces from back to front
        def distance(i):
            v_rotated = q.rotate([facelet_x[i], facelet_y[i], facelet_z[i]])
            return np.linalg.norm(v_rotated - self.camera)

        faces = [(range(9 * i, 9 * (i + 1)), distance(9 * i + 4)) for i in range(0, 6)]
        faces.sort(key=lambda x: -x[1])
        faces = [f for f, d in faces]

        hidden_color = self.get_palette().hidden_color
        for f, face in enumerate(faces):
            for i in face:
                color = self.colors[i]
                if self.hide_nearest_faces and f >= 3:
                    color = hidden_color
                if f < 3 and color != hidden_color:
                    color = color[:3] + (255,)
                self.draw_facelet(
                    painter,
                    w,
                    h,
                    rotation_matrix,
                    facelet_x[i],
                    facelet_y[i],
                    facelet_z[i],
                    color,
                    FACELET_AXIS[i],
                )

    def rotate(self, dx, dy=0):
        self.view_y += dx * 0.005
        self.view_x += dy * 0.005


def rotation_for(o: Orientation) -> Quaternion:
    # Return the quaternion that brings a default cube into the given orientation
    base = Orientation("u", "f")
    if o.top in "fb":
        ticks = AXIS_ROTATIONS["r"].index(o.top)
        base = base.x(ticks)
        q = Quaternion(axis=[1, 0, 0], angle=-math.pi / 2 * ticks)
    else:
        ticks = AXIS_ROTATIONS["f"].index(o.top)
        q = Quaternion(axis=[0, 1, 0], angle=math.pi / 2 * ticks)
        base = base.z(ticks)
    ticks = (
        AXIS_ROTATIONS[base.top].index(o.front)
        - AXIS_ROTATIONS[base.top].index(base.front)
        + 4
    )
    q = Quaternion(axis=[0, 0, 1], angle=-math.pi / 2 * ticks) * q
    return q


class CubeWidget(QWidget):
    """OpenGL widget that uses CubeViz for rendering"""

    def __init__(self, viz: CubeViz, parent=None):
        super(CubeWidget, self).__init__(parent)
        self.setAttribute(Qt.WA_TranslucentBackground)

        # Timer for update loop
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_surface)
        self.timer.start(30)  # ~60 FPS

        self.setMinimumSize(400, 400)

        self.viz = viz
        self.viz.attempt.add_cube_listener(self.refresh)
        self.previous_solution = self.viz.attempt.solution

        # Mouse tracking
        self.setMouseTracking(True)
        self.last_mouse_pos = None
        self.dragging = False

    def refresh(self):
        # Repaint
        self.update()

    def update_surface(self):
        self.update()

    def paintEvent(self, event):
        painter = QPainter(self)
        self.viz.draw(painter, self.width(), self.height())

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.dragging = True
            self.last_mouse_pos = event.pos()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.dragging = False

    def mouseMoveEvent(self, event):
        if self.dragging and self.last_mouse_pos:
            dx = event.x() - self.last_mouse_pos.x()
            dy = event.y() - self.last_mouse_pos.y()
            self.viz.rotate(dx, dy)
            self.last_mouse_pos = event.pos()

    def export_png(self, file_path, size=None):
        """Export the cube widget as a PNG image with transparent background.

        Args:
            file_path (str): Path where the PNG file will be saved
        """
        # Create a pixmap with transparent background
        w, h = self.width(), self.height()
        if size:
            w, h = size, size
        pixmap = QPixmap(QSize(w, h))
        pixmap.fill(Qt.transparent)

        # Paint the widget onto the pixmap
        painter = QPainter(pixmap)
        self.viz.draw(painter, w, h, fill_bg=False)
        painter.end()

        # Save the pixmap to file
        pixmap.save(file_path, "PNG")



import json
import logging
import os
import sys
from dataclasses import dataclass, field, asdict
from functools import cached_property
from pathlib import Path
from typing import List, Tuple

from PyQt5.QtWidgets import (
    QDialog,
    QWidget,
    QPushButton,
    QVBoxLayout,
    QHBoxLayout,
    QGroupBox,
    QCheckBox,
    QDialogButtonBox,
    QSlider,
)
from PyQt5.QtCore import Qt

# Factory-default cube face colors
_DEFAULT_COLORS = [
    (255, 255, 255),
    (255, 255, 0),
    (0, 153, 0),
    (0, 0, 255),
    (255, 0, 0),
    (255, 94, 51),  # (255, 204, 25),
]


class RecognitionOptionNames:
    BAD_FACES = "bad faces"
    BAD_PIECES = "bad pieces"
    TOP_COLOR = "top color"
    BOTTOM_COLOR = "bottom color"
    ALL = "all"


@dataclass
class RecognitionOptions:
    """Preference setting for drawing the cube while solving different steps"""

    eo_edges: List[str]
    eo_corners: List[str]
    dr_edges: List[str]
    dr_corners: List[str]
    htr_edges: List[str]
    htr_corners: List[str]
    fr_edges: List[str]
    fr_corners: List[str]

    @staticmethod
    def default() -> "RecognitionOptions":
        return RecognitionOptions(
            eo_edges=[RecognitionOptionNames.BAD_PIECES],
            eo_corners=[],
            dr_edges=[
                RecognitionOptionNames.BAD_FACES,
            ],
            dr_corners=[RecognitionOptionNames.BAD_FACES],
            htr_edges=[RecognitionOptionNames.BAD_FACES],
            htr_corners=[
                RecognitionOptionNames.BAD_FACES,
                RecognitionOptionNames.BOTTOM_COLOR,
            ],
            fr_edges=[RecognitionOptionNames.BAD_FACES],
            fr_corners=[RecognitionOptionNames.BAD_FACES],
        )

    @staticmethod
    def minimal() -> "RecognitionOptions":
        return RecognitionOptions(
            eo_edges=[RecognitionOptionNames.BAD_PIECES],
            eo_corners=[],
            dr_edges=[RecognitionOptionNames.BAD_FACES],
            dr_corners=[RecognitionOptionNames.BAD_FACES],
            htr_edges=[RecognitionOptionNames.BAD_FACES],
            htr_corners=[RecognitionOptionNames.BAD_FACES],
            fr_edges=[RecognitionOptionNames.BAD_FACES],
            fr_corners=[RecognitionOptionNames.BAD_FACES],
        )


@dataclass
class Preferences:
    """Top-level preferences object"""

    opacity: int = 237
    background_color: int = 77
    colors: List[Tuple] = field(default_factory=lambda: _DEFAULT_COLORS)
    recognition: RecognitionOptions = field(default_factory=RecognitionOptions.default)
    listeners: List = field(default_factory=list)

    def save(self):
        prefs_path = Preferences.get_preferences_path()

        prefs = {
            "opacity": self.opacity,
            "recognition": asdict(self.recognition),
            "colors": self.colors,
            "background": self.background_color,
        }

        try:
            with open(prefs_path, "w") as f:
                json.dump(prefs, f)
        except Exception as e:
            logging.error(f"Error saving preferences: {e}")

    def add_listener(self, callback):
        self.listeners.append(callback)

    def notify(self):
        for listener in self.listeners:
            listener()

    @staticmethod
    def load() -> "Preferences":
        # Get the preferences file path
        prefs_path = Preferences.get_preferences_path()

        if prefs_path.exists():
            try:
                with open(prefs_path, "r") as f:
                    prefs = json.load(f)
                    recognition = asdict(RecognitionOptions.default())
                    recognition.update(prefs.get("recognition", {}))
                    opacity = prefs.get("opacity", 237)
                    colors = prefs.get("colors", [])
                    if len(colors) != len(_DEFAULT_COLORS):
                        colors = _DEFAULT_COLORS
                    bg = prefs.get("background", 77)
                    return Preferences(
                        opacity=opacity,
                        colors=colors,
                        recognition=RecognitionOptions(**recognition),
                        background_color=bg,
                    )
            except Exception as e:
                logging.error(f"Error loading preferences: {e}")
                return Preferences()
        else:
            return Preferences()

    @staticmethod
    def get_preferences_path() -> Path:
        return app_dir() / "preferences.json"


class PreferencesDialog(QDialog):
    """Dialog for modifying preferences"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Preferences")
        self.setWindowModality(Qt.NonModal)
        self.setMinimumWidth(300)

        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.colors_widget)
        layout.addWidget(self.eo_widget)
        layout.addWidget(self.dr_widget)
        layout.addWidget(self.htr_widget)
        layout.addWidget(self.fr_widget)

        def close():
            preferences.save()
            self.hide()

        button_box = QDialogButtonBox(QDialogButtonBox.Save)
        button_box.accepted.connect(close)
        layout.addWidget(button_box)

    def piece_options(
        self, name: str, target_list: List[str], options: List[str], forced: List[str]
    ) -> QCheckBox:
        boxes = []

        def update():
            target_list.clear()
            for b in boxes:
                if b.isChecked():
                    target_list.append(b.objectName())
            preferences.notify()

        group = QGroupBox(name)
        layout = QHBoxLayout()
        group.setLayout(layout)
        for option in options:
            box = QCheckBox(option)
            box.setObjectName(option)
            box.setChecked(option in target_list or option in forced)
            box.setEnabled(option not in forced)
            box.stateChanged.connect(update)
            boxes.append(box)
            layout.addWidget(box)
        layout.addWidget(QWidget(), 1)
        return group

    def step_options(
        self,
        name: str,
        edge_target: List[str],
        edge_options: List[str],
        edge_disabled: List[str],
        corner_target: List[str],
        corner_options: List[str],
        corner_disabled: List[str],
    ) -> QGroupBox:
        group = QGroupBox(name)
        layout = QHBoxLayout()
        group.setLayout(layout)
        layout.addWidget(
            self.piece_options("Edges", edge_target, edge_options, edge_disabled)
        )
        layout.addWidget(
            self.piece_options(
                "Corners", corner_target, corner_options, corner_disabled
            )
        )
        return group

    @cached_property
    def colors_widget(self) -> QWidget:
        group = QGroupBox("Colors")
        layout = QHBoxLayout()
        group.setLayout(layout)
        layout.addWidget(self.cube_colors_widget)
        layout.addWidget(self.opacity_widget)
        layout.addWidget(self.background_widget)
        return group

    @cached_property
    def cube_colors_widget(self) -> QWidget:
        group = QGroupBox("Cube")
        layout = QHBoxLayout()
        group.setLayout(layout)

        col = [QVBoxLayout(), QVBoxLayout(), QVBoxLayout()]
        for c in col:
            layout.addLayout(c)

        color_names = ["U", "D", "F", "B", "R", "L"]
        color_buttons = []
        for i, color in enumerate(preferences.colors):
            color_button = QWidget()
            color_button.setFixedSize(30, 30)
            color_button.setStyleSheet(
                f"background-color: rgb({color[0]}, {color[1]}, {color[2]}); border: 1px solid black;"
            )
            color_button.setCursor(Qt.PointingHandCursor)
            color_button.setToolTip(color_names[i])

            # Store the index for the clicked callback
            color_button.index = i

            # Connect mouse press event via installEventFilter
            color_button.mousePressEvent = lambda event, idx=i: self._show_color_dialog(
                idx
            )

            color_buttons.append(color_button)
            col[int(i / 2)].addWidget(color_button)

        def reset():
            preferences.colors = _DEFAULT_COLORS
            for i, c in enumerate(preferences.colors):
                color_buttons[i].setStyleSheet(
                    f"background-color: rgb({c[0]}, {c[1]}, {c[2]}); border: 1px solid black;"
                )
            preferences.notify()

        reset_button = QPushButton("Reset")
        layout.addWidget(reset_button)
        reset_button.clicked.connect(reset)

        layout.addWidget(self.background_widget)

        layout.addStretch(1)

        return group

    def _show_color_dialog(self, index):
        """Show a color dialog for selecting the color at the specified index"""
        from PyQt5.QtWidgets import QColorDialog
        from PyQt5.QtGui import QColor

        # Store a reference to self inside the color picker
        # This helps maintain the parent relationship
        self._current_color_picker = QColorDialog(self)

        # Set up the dialog with current color
        current_color = preferences.colors[index]
        initial_color = QColor(current_color[0], current_color[1], current_color[2])
        self._current_color_picker.setCurrentColor(initial_color)
        self._current_color_picker.setWindowTitle(f"Select Color {index+1}")

        # Define callback for when a color is selected
        def on_color_selected(color):
            if color.isValid():
                # Update the color in preferences
                new_color = (color.red(), color.green(), color.blue())
                preferences.colors[index] = new_color

                # Update the button appearance
                for button in self.findChildren(QWidget):
                    if hasattr(button, "index") and button.index == index:
                        button.setStyleSheet(
                            f"background-color: rgb({new_color[0]}, {new_color[1]}, {new_color[2]}); border: 1px solid black;"
                        )
                        break

                # Notify listeners about the change
                preferences.notify()

        # Connect signal and show the dialog
        self._current_color_picker.colorSelected.connect(on_color_selected)
        self._current_color_picker.finished.connect(lambda: self.activateWindow())
        self._current_color_picker.setModal(False)  # Make it non-modal
        self._current_color_picker.show()

    @cached_property
    def eo_widget(self) -> QWidget:
        minimal = RecognitionOptions.minimal()
        return self.step_options(
            "EO Recognition",
            preferences.recognition.eo_edges,
            [RecognitionOptionNames.BAD_PIECES, RecognitionOptionNames.ALL],
            minimal.eo_edges,
            preferences.recognition.eo_corners,
            [RecognitionOptionNames.ALL],
            minimal.eo_corners,
        )

    @cached_property
    def dr_widget(self) -> QWidget:
        minimal = RecognitionOptions.minimal()
        return self.step_options(
            "DR Recognition",
            preferences.recognition.dr_edges,
            [
                RecognitionOptionNames.BAD_FACES,
                RecognitionOptionNames.BAD_PIECES,
                RecognitionOptionNames.ALL,
            ],
            minimal.dr_edges,
            preferences.recognition.dr_corners,
            [
                RecognitionOptionNames.BAD_FACES,
                RecognitionOptionNames.BAD_PIECES,
                RecognitionOptionNames.ALL,
            ],
            minimal.dr_corners,
        )

    @cached_property
    def htr_widget(self) -> QWidget:
        minimal = RecognitionOptions.minimal()
        return self.step_options(
            "HTR Recognition",
            preferences.recognition.htr_edges,
            [
                RecognitionOptionNames.BAD_FACES,
                RecognitionOptionNames.BAD_PIECES,
                RecognitionOptionNames.TOP_COLOR,
                RecognitionOptionNames.BOTTOM_COLOR,
                RecognitionOptionNames.ALL,
            ],
            minimal.htr_edges,
            preferences.recognition.htr_corners,
            [
                RecognitionOptionNames.BAD_FACES,
                RecognitionOptionNames.BAD_PIECES,
                RecognitionOptionNames.TOP_COLOR,
                RecognitionOptionNames.BOTTOM_COLOR,
                RecognitionOptionNames.ALL,
            ],
            minimal.htr_corners,
        )

    @cached_property
    def fr_widget(self) -> QWidget:
        minimal = RecognitionOptions.minimal()
        return self.step_options(
            "FR Recognition",
            preferences.recognition.fr_edges,
            [
                RecognitionOptionNames.BAD_FACES,
                RecognitionOptionNames.BAD_PIECES,
                RecognitionOptionNames.ALL,
            ],
            minimal.fr_edges,
            preferences.recognition.fr_corners,
            [
                RecognitionOptionNames.BAD_FACES,
                RecognitionOptionNames.BAD_PIECES,
                RecognitionOptionNames.ALL,
            ],
            minimal.fr_corners,
        )

    @cached_property
    def background_widget(self) -> QWidget:
        group = QGroupBox("Background Color")
        layout = QHBoxLayout()
        group.setLayout(layout)
        slider = QSlider(Qt.Horizontal)
        layout.addWidget(slider)
        layout.addStretch(1)
        slider.setMinimum(0)
        slider.setMaximum(255)
        slider.setValue(preferences.background_color)

        def update():
            preferences.background_color = slider.value()
            preferences.notify()

        slider.valueChanged.connect(update)

        return group

    @cached_property
    def opacity_widget(self) -> QWidget:
        group = QGroupBox("Transparency")
        layout = QHBoxLayout()
        group.setLayout(layout)
        slider = QSlider(Qt.Horizontal)
        layout.addWidget(slider)
        layout.addStretch(1)
        slider.setMinimum(0)
        slider.setMaximum(75)
        slider.setValue(255 - preferences.opacity)

        def update():
            preferences.opacity = 255 - slider.value()
            preferences.notify()

        slider.valueChanged.connect(update)

        return group


_dialog = None


def show_dialog(parent):
    """Show preferences dialog and apply settings if accepted"""
    global _dialog
    if _dialog is None or not _dialog.isVisible():
        # Create a new dialog if it doesn't exist or isn't visible
        _dialog = PreferencesDialog(parent)
    _dialog.show()
    _dialog.raise_()  # Bring to front
    _dialog.activateWindow()  # Set as active window


def app_dir() -> Path:
    """Return platform-appropriate application home directory"""
    if sys.platform == "darwin":  # macOS
        path = Path.home() / "Library" / "Preferences" / "vfmc"
    elif sys.platform == "win32":  # Windows
        path = Path(os.environ.get("APPDATA", str(Path.home()))) / "vfmc"
    else:  # Linux/Unix
        path = Path.home() / ".config" / "vfmc"
    path.mkdir(parents=True, exist_ok=True)
    return path


# Preferences as saved by the user
preferences = Preferences.load()



from enum import Enum
from typing import Dict, Tuple

from vfmc.prefs import preferences, RecognitionOptionNames


class Visibility:
    """Bit mask to describe facelets while solving a step"""

    Any = 1
    BadFace = 2
    BadPiece = 4
    BottomColor = 8
    TopColor = 16
    All = 255


# Map of preference keys to Visibility bit mask
_VISIBILITY_NAMES = {
    RecognitionOptionNames.BAD_FACES: Visibility.BadFace,
    RecognitionOptionNames.BAD_PIECES: Visibility.BadPiece,
    RecognitionOptionNames.TOP_COLOR: Visibility.TopColor,
    RecognitionOptionNames.BOTTOM_COLOR: Visibility.BottomColor,
    RecognitionOptionNames.ALL: Visibility.All,
}


class FaceletColors(Enum):
    WHITE = 0
    YELLOW = 1
    GREEN = 2
    BLUE = 3
    RED = 4
    ORANGE = 5


class Palette:
    def __init__(
        self,
        colors: Dict[FaceletColors, Tuple],
        edge_visibility_mask: int,
        center_visibility_mask: int,
        corner_visibility_mask: int,
        hidden_color: Tuple,
        opacity: int,
    ):
        """Colors for drawing the cube"""
        self.colors = colors
        self.edge_visibility_mask = edge_visibility_mask
        self.center_visibility_mask = center_visibility_mask
        self.corner_visibility_mask = corner_visibility_mask
        self.hidden_color = hidden_color
        self.opacity = opacity

    def color_of_edge(self, f: FaceletColors, visibility: int) -> Tuple:
        if visibility & self.edge_visibility_mask == 0:
            return self.hidden_color
        c = self.colors[f]
        return self.colors[f] + (self.opacity,) if len(c) < 4 else c

    def color_of_center(self, f: FaceletColors, visibility: int) -> Tuple:
        if visibility & self.center_visibility_mask == 0:
            return self.hidden_color
        c = self.colors[f]
        return self.colors[f] + (self.opacity,) if len(c) < 4 else c

    def color_of_corner(self, f: FaceletColors, visibility: int) -> Tuple:
        if visibility & self.corner_visibility_mask == 0:
            return self.hidden_color
        c = self.colors[f]
        return self.colors[f] + (self.opacity,) if len(c) < 4 else c

    @staticmethod
    def by_name(name) -> "Palette":
        colors = dict(
            (FaceletColors(k), tuple(v)) for k, v in enumerate(preferences.colors)
        )
        bg = preferences.background_color
        h, o = (0, 51) if bg > 128 else (255, 51)
        hidden = (h, h, h, o)
        p = Palette(colors, 0, 255, 0, hidden, preferences.opacity)
        options = [], []
        if name == "eo":
            options = (
                preferences.recognition.eo_edges,
                preferences.recognition.eo_corners,
            )
        elif name == "dr":
            options = (
                preferences.recognition.dr_edges,
                preferences.recognition.dr_corners,
            )
        elif name == "htr":
            options = (
                preferences.recognition.htr_edges,
                preferences.recognition.htr_corners,
            )
        elif name == "fr":
            options = (
                preferences.recognition.fr_edges,
                preferences.recognition.fr_corners,
            )
        elif name == "finish":
            p.edge_visibility_mask = Visibility.All
            p.corner_visibility_mask = Visibility.All
        elif name == "insertions":
            p.edge_visibility_mask = Visibility.BadPiece
            p.corner_visibility_mask = Visibility.BadPiece
        elif name == "cp-case":
            p.center_visibility_mask = 0
            p.opacity = 255
            p.corner_visibility_mask = Visibility.BadPiece
            p.colors = dict((FaceletColors(i), (0, 0, 0)) for i in range(6))
            p.colors.update(
                {
                    FaceletColors.BLUE: p.hidden_color,
                    FaceletColors.GREEN: p.hidden_color,
                }
            )
        elif name == "co-case":
            p.center_visibility_mask = 0
            p.opacity = 255
            p.corner_visibility_mask = Visibility.BadFace
            p.colors[FaceletColors.ORANGE] = p.colors[FaceletColors.RED]
            p.colors[FaceletColors.BLUE] = p.colors[FaceletColors.GREEN]
            p.colors[FaceletColors.YELLOW] = p.colors[FaceletColors.WHITE]
        elif name == "htr-corner-case":
            p.center_visibility_mask = 0
            p.opacity = 255
            p.corner_visibility_mask = Visibility.BadFace
            p.colors[FaceletColors.ORANGE] = p.colors[FaceletColors.RED]
            p.colors[FaceletColors.BLUE] = p.colors[FaceletColors.GREEN]
        elif name == "htr-case":
            p.center_visibility_mask = 0
            p.opacity = 255
            p.edge_visibility_mask = Visibility.BadFace
            p.colors[FaceletColors.ORANGE] = p.colors[FaceletColors.RED]
            p.colors[FaceletColors.BLUE] = p.colors[FaceletColors.GREEN]
        elif name == "corner-edge-case":
            p.center_visibility_mask = 0
            p.opacity = 255
            p.corner_visibility_mask = Visibility.BadFace
            p.edge_visibility_mask = Visibility.BadFace
            p.colors[FaceletColors.ORANGE] = p.colors[FaceletColors.RED]
            p.colors[FaceletColors.BLUE] = p.colors[FaceletColors.GREEN]
        else:
            p.edge_visibility_mask = Visibility.All
            p.corner_visibility_mask = Visibility.All
        for opt in options[0]:
            p.edge_visibility_mask |= _VISIBILITY_NAMES.get(opt, 0)
        for opt in options[1]:
            p.corner_visibility_mask |= _VISIBILITY_NAMES.get(opt, 0)
        return p
